#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
.. currentmodule:: taurus.core.evaluation

Evaluation extension for taurus core model.

The evaluation extension is a special extension that provides evaluation
objects. The official scheme name is 'eval'.

The main usage for this extension is to provide a way of performing mathematical
evaluations with data from other source, as well as allowing fast interfacing
with sources of data not supported by specific schemes.

The Evaluation Factory (:class:`EvaluationFactory`) uses the following object
naming for referring to attributes (:class:`EvaluationAttribute`):

    `eval:[//<authority>][@<evaluator>/][<subst>;]<expr>`

or the following for referring to evaluation devices
(:class:`EvaluationDevice`):

    `eval:[//<authority>]@<evaluator>`

or the following for referring to an evaluation authority
(:class:`EvaluationAuthority`):

    `eval://<authority>`

where:

    - The `<authority>` segment is optional (except when referring to an
      EvaluationAuthority). At this point, only `//localhost` is supported.

    - The `@<evaluator>` is optional (except when referring to an
      EvaluationDevice). If not given, it defaults to `DefaultEvaluator`. See
      below for further details

    - `<expr>` is a mathematical expression (using python syntax)
      that may have references to other taurus **attributes** by enclosing them
      between `{` and `}`. Expressions will be evaluated by the evaluator device
      to which the attribute is assigned.

    - The optional `<subst>` segment is used to provide substitution symbols.
      `<subst>` is a semicolon-separated string of `<key>=<value>` strings.

The evaluator device inherits from :class:`SafeEvaluator` which by default
includes a large subset of mathematical functions from the :mod:`numpy`
module. If access to other symbols are required, a custom evaluator can
be used. `<evaluator>` is a unique identification name for the evaluator
device object and may define a source of additional symbols to be present
for the evaluation. The supported syntax for `@<evaluator>` is:

    - `@<ID>` (cannot contain dots or any of `/` `?` `#` `:` `=`). This
      indicates just an alternative name for the EvaluationDevice, It does not
      add any extra symbol to the evaluation context.

    - `@<modulename>.*` (<modulename> may include dots for submodules). It
      will make all symbols found in the given module available during the
      evaluation (i.e., it emulates doing `from <modulename> import *` in the
      evaluation context).

    - `@<modulename>.<customdeviceclass>`. Use your own custom EvaluationDevice
      based class. This allows to define custom symbols see
      :file:`<taurus>/core/evaluation/test/res/dev_example.py`, **but** note
      that this syntax is is now superseded by the "instance-based" one
      (see below), which is easier to use and provides write attribute support.

    - `@<inst>=<modulename>.<class>()` (e.g. `@c=mymod.MyClass()` ). This
      will import a class from a module, then instantiate it and then
      make the instance available for evaluation with the given name. Note that
      the `<inst>=` part may be omitted, in which case the instance will be
      available for evaluation as `self`. **IMPORTANT:** If the given class
      declares writable properties, EvaluationAttributes that access one such
      property will automatically be considered writable. See examples of usage
      in :file:`<taurus>/core/evaluation/test/res/mymod.py` and in
      :file:`<taurus>/core/evaluation/test/res/ipap_example.py`


Some examples of valid evaluation models are:

    - An attribute that multiplies a tango attribute by 2::

        eval:2*{tango:a/b/c/d}

    - Same as above, but using substitutions::

        eval:k=2;a={tango:a/b/c/d};k*a

    - An attribute that adds two tango attributes together (assuming that tango
      is set as the default scheme)::

        eval:{a/b/c/d}+{f/g/h/i}

    - An attribute that generates an array of random values::

        eval:rand(256)

    - Same as above, but with units::

        eval:Q(rand(256),'V')

    - An attribute that adds noise to a tango image attribute::

        eval:img={tango:sys/tg_test/1/short_image_ro};img+10*rand(*img.shape)

    - An attribute that accesses a method from a given module (in this
      case to use os.path.exists)::

        eval:@os.*/path.exists("/some/file")

    - Same as before, for getting today's date as an attribute::

        eval:@datetime.*/date.today().isoformat()

    - A default evaluator device named `foo`::

        eval:@foo

    - A custom evaluator device (implemented as class `MyEvalDev` in the `mymod`
      module)::

        eval:@mymod.MyEvalDev

    - A custom evaluator device (implemented as class `MyEvalDev` in the `mymod`
      module)::

        eval:@mymod.MyEvalDev

    - A writable attribute foo (implemented as a writable property of the
      `MyClass` class from the `mymod` module)::

        eval:@c=mymod.MyClass()/c.foo

      assuming that the `mymod` module defines `MyClass` as::

        class MyClass(object):
            (...)
            get_foo(self):
                (...)
            set_foo(self, value):
                (...)
            foo = property(get_foo, set_foo)
            (...)


.. note:: Previous to SEP3, a RFC3986 non-compliant syntax was used for the
          evaluation scheme (e.g., allowing names such as
          ``eval://db=foo;dev=bar;a*b?k=2;a={tango:a/b/c/d}``).
          This syntax is now deprecated and should not be used. Taurus will
          issue warnings if detected.
"""
from __future__ import absolute_import

from .evalfactory import EvaluationFactory
from .evalattribute import EvaluationAttribute
from .evalauthority import EvaluationAuthority
from .evaldevice import EvaluationDevice



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

from taurus.core.units import Q_

# from pint import UnitRegistry
# UR = UnitRegistry()
# Q_= UR.Quantity
# UR.define(u'degreeC = kelvin; offset: 273.15  = °C = ºC = degC = celsius')


def test_celsius():
    """
    Check that the definition of degreeC works
    Note: see https://github.com/hgrecco/pint/issues/546
    """
    a = Q_(3, u"°C")
    b = Q_(3, u"ºC")
    c = Q_(3, u"degC")
    d = Q_(3, u"celsius")
    e = Q_(276.15, u"kelvin")
    assert(a == b)
    assert(a == c)
    assert(a == d)
    assert(a == e)
    assert(a.to(u"kelvin") == e)
    assert(a == e.to(u"°C"))
    assert(a == e.to(u"ºC"))
    assert(a == e.to(u"celsius"))
    assert(u"{}".format(a) == u"3 °C")
    assert(u"{}".format(b) == u"3 °C")
    assert(u"{}".format(c) == u"3 °C")
    assert(u"{}".format(d) == u"3 °C")

    # # I would have expected the following to be formatted as "3 celsius"...
    # a.default_format = u""
    # b.default_format = u""
    # c.default_format = u""
    # d.default_format = u""
    # assert(u"{}".format(a) == u"3 celsius")
    # assert(u"{}".format(b) == u"3 celsius")
    # assert(u"{}".format(c) == u"3 celsius")
    # assert(u"{}".format(d) == u"3 celsius")



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides base class for all taurus container widgets"""

__all__ = ["TaurusBaseContainer"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt

from taurus.qt.qtgui.base import TaurusBaseComponent, TaurusBaseWidget


class TaurusBaseContainer(TaurusBaseWidget):
    """Base class for the Taurus container widgets.
    This type of taurus container classes are specially useful if you define
    a parent taurus model to them and set all contained taurus widgets to use parent
    model. Example::

        from taurus.qt.qtgui.container import *
        from taurus.qt.qtgui.display import *

        widget = TaurusWidget()
        layout = Qt.QVBoxLayout()
        widget.setLayout(layout)
        widget.model = 'sys/database/2'
        stateWidget = TaurusLabel()
        layout.addWidget(stateWidget)
        stateWidget.model = 'sys/database/2/state'
    """

    def __init__(self, name='', parent=None, designMode=False):
        name = name or self.__class__.__name__

        self.call__init__(TaurusBaseWidget, name,
                          parent, designMode=designMode)

        self.defineStyle()
        self.designMode = designMode

    def taurusChildren(self, objs=None):
        '''
        returns a list of all taurus children of this taurus container (recurses down
        skipping non-taurus widgets)

        :param objs: (list<objects>) if given, the search starts at the objects
                     passed in this list

        :return: (list<TaurusBaseWidget>)
        '''
        if objs is None:
            objs = self.children()
        result = []
        for o in objs:
            if isinstance(o, TaurusBaseWidget):
                result.append(o)
            else:
                result += self.taurusChildren(o.children())
        return result

    def defineStyle(self):
        self.updateStyle()

    def sizeHint(self):
        #        if self.designMode:
        #            return Qt.QSize(150, 150)
        return Qt.QWidget.sizeHint(self)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget over writing
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def isReadOnly(self):
        return True

    def updateStyle(self):
        if self.getShowQuality():
            self.setAutoFillBackground(True)
            # TODO: get quality/state from model and update accordingly
        else:
            self.setAutoFillBackground(False)
            # TODO: restore colors
        TaurusBaseWidget.updateStyle(self)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Pending operations related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getPendingOperations(self):
        pending_ops = []

        for child in Qt.QObject.children(self):
            if isinstance(child, TaurusBaseComponent):
                pending_ops += child.getPendingOperations()
        return pending_ops

    def resetPendingOperations(self):
        self.debug("Reset changes")
        for child in Qt.QObject.children(self):
            if isinstance(child, TaurusBaseComponent):
                child.resetPendingOperations()

    def hasPendingOperations(self):
        ret = False
        for child in Qt.QObject.children(self):
            if isinstance(child, TaurusBaseComponent):
                ret |= child.hasPendingOperations()
        return ret

    def handleEvent(self, evt_src, evt_type, evt_value):
        if not self._setText or not self.getShowText():
            return
        modelObj = self.getModelObj()
        if modelObj:
            txt = modelObj.getDisplayName(complete=False)
        else:
            txt = self.getNoneValue()
        self._setText(txt)

#    @classmethod
#    def getQtDesignerPluginInfo(cls):
#        """Returns pertinent information in order to be able to build a valid
#        QtDesigner widget plugin
#
#        The dictionary returned by this method should contain *at least* the
#        following keys and values:
#        - 'module' : a string representing the full python module name (ex.: 'taurus.qt.qtgui.base')
#        - 'icon' : a string representing valid resource icon (ex.: 'designer:combobox.png')
#        - 'container' : a bool telling if this widget is a container widget or not.
#
#        This default implementation returns the following dictionary:
#            { 'group'     : 'Taurus Containers',
#              'icon'      : 'designer:widget.png',
#              'container' : True }
#
#        :return: (dict) a map with pertinent designer information"""
#        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
#        ret['module'] = 'taurus.qt.qtgui.container'
#        ret['group'] = 'Taurus Containers'
#        ret['icon'] = "designer:widget.png"
#        ret['container'] = True
#        return ret



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains the taurus base listeners classes"""

from __future__ import print_function
from builtins import object
from .util.log import Logger


__all__ = ["TaurusListener", "TaurusExceptionListener"]

__docformat__ = "restructuredtext"


class TaurusListener(Logger):
    """ TaurusListener Interface"""

    def __init__(self, name='', parent=None):
        self.call__init__(Logger, name, parent)

    def eventReceived(self, src, type, evt_value):
        """ Method to implement the event notification"""
        pass

    def attributeList(self):
        """ Method to return the attributes of the widget"""
        return []


class TaurusExceptionListener(object):
    """Class for handling ConnectionFailed, DevFailed and TaurusException exceptions."""

    def connectionFailed(self, exception):
        msg = 'Deprecation warning: please note that the "connectionFailed" ' +\
              'method is deprecated. Scheme-specific exceptions should be ' +\
              'implemented in each model and be transformed into taurus ' +\
              'exceptions according Sep3 specifications'
        self.info(msg)
        self._printException(exception)

    def devFailed(self, exception):
        msg = 'Deprecation warning: please note that the "devFailed" ' +\
              'method is deprecated. Scheme-specific exceptions should be ' +\
              'implemented in each model and be transformed into taurus ' +\
              'exception according Sep3 specifications'
        self.info(msg)
        self._printException(exception)

    def exceptionReceived(self, exception):
        self._printException(exception)

    def _printException(self, exception):
        print(self.__class__.__name__, "received", exception.__class__.__name__, str(exception))



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
  This module provides the :class:`TaurusFactory` base class that any valid
  Factory in Taurus must inherit.

  The Factory objects are the basic block for building and interacting with a
  given scheme in Taurus. They provide Taurus model objects (TaurusAuthority,
  TaurusDevice or TaurusAttribute) for a given taurus model name.

  Taurus model naming is URI based (see <https://tools.ietf.org/html/rfc3986>)

  All the standard components of an URI (scheme, authority, path, query and
  fragment) may be part of a model name, and they are separated as follows:

  <scheme>:<authority><path>?<query>#<fragment>


  The following are some points to consider when using and/or implementing
  schemes based on this Abstract class:

  - It is strongly recommended that the scheme component is always present
  explicitly in the model name, although a default scheme can be defined in
  :mod:`taurus.tauruscustomsettings` so that model names which do not explicit
  the scheme can be auto-completed.

  - The authority component (if present on a given name) must always begin by
  a double slash ('//'). (see
  <https://tools.ietf.org/html/rfc3986#section-3.2>)

  - The path component, if present, must start by a single slash ('/') (see
  <https://tools.ietf.org/html/rfc3986#section-3.3>)

"""
from __future__ import absolute_import
from builtins import object

import atexit
from weakref import WeakValueDictionary
from .taurusbasetypes import TaurusElementType
from .taurusauthority import TaurusAuthority
from .taurusdevice import TaurusDevice
from .taurusattribute import TaurusAttribute
from .taurusconfiguration import TaurusConfiguration, TaurusConfigurationProxy
from .taurusexception import TaurusException
from taurus.core.tauruspollingtimer import TaurusPollingTimer

__all__ = ["TaurusFactory"]

__docformat__ = "restructuredtext"


class TaurusFactory(object):
    """The base class for valid Factories in Taurus."""

    schemes = ()  # reimplement in derived classes to declare supported schemes
    caseSensitive = True  # reimplement if your scheme is case insensitive

    elementTypesMap = None  # reimplement in derived classes to profit from
                            # generic implementations of getAuthority,
                            # getDevice, getAttribute, findObjectClass, etc.
                            # see findObjectClass for more details

    DefaultPollingPeriod = 3000

    def __init__(self):
        atexit.register(self.cleanUp)
        self._polling_period = self.DefaultPollingPeriod
        self.polling_timers = {}
        self._polling_enabled = True
        self._attrs = WeakValueDictionary()
        self._devs = WeakValueDictionary()
        self._auths = WeakValueDictionary()

        from . import taurusmanager
        manager = taurusmanager.TaurusManager()
        self._serialization_mode = manager.getSerializationMode()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for cleanUp at exit
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def cleanUp(self):
        """Reimplement if you need to execute code on program execution exit.
        Default implementation does nothing.
        """
        pass

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for serialization
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setSerializationMode(self, mode):
        """Sets the serialization mode for the system.

        :param mode: (TaurusSerializationMode) the new serialization mode
        """
        self._serialization_mode = mode

    def getSerializationMode(self):
        """Gives the serialization operation mode.

        :return: (TaurusSerializationMode) the current serialization mode
        """
        return self._serialization_mode

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API to get objects. Generic implementation. You may want to reimplement
    # it in your scheme factory
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getAuthority(self, name=None):
        """Obtain the model object corresponding to the given authority name.
        If the corresponding authority already exists, the existing instance
        is returned. Otherwise a new instance is stored and returned.

        :param name: (str) authority name

        :return: a taurus.core.taurusauthority.TaurusAuthority object
        :raises: :TaurusException: if the given name is invalid.
        """
        v = self.getAuthorityNameValidator()
        if not v.isValid(name):
            msg = "Invalid {scheme} authority name '{name}'".format(
                    scheme=self.schemes[0], name=name)
            raise TaurusException(msg)

        fullname, _, _ = v.getNames(name)
        auth = self._auths.get(fullname)
        if auth is not None:
            return auth

        cls = self.elementTypesMap[TaurusElementType.Authority]
        auth = cls(name=fullname)
        self._auths[fullname] = auth
        return auth

    def getDevice(self, name, **kw):
        """Obtain the model object corresponding to the given device name.
        If the corresponding device already exists, the existing instance
        is returned. Otherwise a new instance is stored and returned.

        :param name: (str) device name

        :return: a taurus.core.taurusdevice.TaurusDevice object
        :raises: :TaurusException: if the given name is invalid.
        """
        v = self.getDeviceNameValidator()
        if not v.isValid(name):
            msg = "Invalid {scheme} device name '{name}'".format(
                    scheme=self.schemes[0], name=name)
            raise TaurusException(msg)

        fullname, _, _ = v.getNames(name)
        dev = self._devs.get(fullname)
        if dev is not None:
            return dev

        try:
            # this works if the authority name is present in the dev full name
            # (which in principle should always be the case)
            authname = v.getUriGroups(fullname)['authority']
            auth = self.getAuthority(authname)
        except:
            self.debug('Cannot get device parent from name "%s"', fullname)
            auth = None

        cls = self.elementTypesMap[TaurusElementType.Device]
        dev = cls(name=fullname, parent=auth)
        self._devs[fullname] = dev
        return dev

    def getAttribute(self, name):
        """ Obtain the model object corresponding to the given attribute name.
        If the corresponding attribute already exists, the existing instance
        is returned. Otherwise a new instance is stored and returned.

        :param name: (str) attribute name

        :return: a taurus.core.taurusattribute.TaurusAttribute object
        :raises: :TaurusException: if the given name is invalid.
        """
        v = self.getAttributeNameValidator()
        if not v.isValid(name):
            msg = "Invalid {scheme} attribute name '{name}'".format(
                    scheme=self.schemes[0], name=name)
            raise TaurusException(msg)

        fullname, _, _ = v.getNames(name)
        attr = self._attrs.get(fullname)
        if attr is not None:
            return attr

        try:
            # this works only if the devname is present in the attr full name
            # (not all schemes are constructed in this way)
            devname = v.getUriGroups(fullname)['devname']
            dev = self.getDevice(devname)
        except:
            self.debug('Cannot get attribute parent from name "%s"', fullname)
            dev = None

        cls = self.elementTypesMap[TaurusElementType.Attribute]
        attr = cls(name=fullname, parent=dev)
        self._attrs[fullname] = attr
        return attr

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Methods that must be implemented by the specific Factory
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getAuthorityNameValidator(self):
        raise NotImplementedError("getAuthorityNameValidator cannot be called"
                                  " for abstract TaurusFactory")

    def getDeviceNameValidator(self):
        raise NotImplementedError("getDeviceNameValidator cannot be called"
                                  " for abstract TaurusFactory")

    def getAttributeNameValidator(self):
        raise NotImplementedError("getAttributeNameValidator cannot be called"
                                  " for abstract TaurusFactory")

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Factory extension API
    # Override the following methods if you need to provide special classes for
    # special object types
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def registerAttributeClass(self, attr_name, attr_klass):
        pass

    def unregisterAttributeClass(self, attr_name):
        pass

    def registerDeviceClass(self, dev_klass_name, dev_klass):
        pass

    def unregisterDeviceClass(self, dev_klass_name):
        pass

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Generic methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def supportsScheme(self, scheme):
        """Returns whether the given scheme is supported by this factory

        :param scheme: (str) the name of the schem to be checked

        :return: (bool) True if the scheme is supported (False otherwise)
        """
        return scheme in self.schemes

    def findObject(self, absolute_name):
        """ Must give an absolute name"""
        if not absolute_name:
            return None
        obj = None
        cls = self.findObjectClass(absolute_name)
        if cls:
            obj = self.getObject(cls, absolute_name)
        return obj

    def getObject(self, cls, name):
        t4_msg = 'The TaurusConfiguration classes are deprecated in tep14'
        if issubclass(cls, TaurusAuthority):
            return self.getAuthority(name)
        elif issubclass(cls, TaurusDevice):
            return self.getDevice(name)
        elif issubclass(cls, TaurusAttribute):
            return self.getAttribute(name)
        # For backward compatibility
        elif issubclass(cls, TaurusConfiguration):
            self.deprecated(dep='TaurusConfiguration', alt='TaurusAttribute',
                            rel='4.0', dbg_msg=t4_msg)
            return self.getAttribute(name)
        elif issubclass(cls, TaurusConfigurationProxy):
            self.deprecated(dep='TaurusConfigurationProxy',
                            alt='TaurusAttribute',
                            rel='4.0', dbg_msg=t4_msg)
            return self.getAttribute(name)
        else:
            return None

    def changeDefaultPollingPeriod(self, period):
        if period > 0:
            self._polling_period = period

    def getDefaultPollingPeriod(self):
        return self._polling_period

    def isPollingEnabled(self):
        """Tells if the Taurus polling is enabled

           :return: (bool) whether or not the polling is enabled
        """
        return self._polling_enabled

    def disablePolling(self):
        """Disable the application tango polling"""
        if not self.isPollingEnabled():
            return
        self._polling_enabled = False
        for period, timer in self.polling_timers.items():
            timer.stop()

    def enablePolling(self):
        """Enable the application tango polling"""
        if self.isPollingEnabled():
            return
        for period, timer in self.polling_timers.items():
            timer.start()
        self._polling_enabled = True

    def addAttributeToPolling(self, attribute, period, unsubscribe_evts=False):
        """Activates the polling (client side) for the given attribute with the
           given period (seconds).

           :param attribute: (taurus.core.taurusattribute.TaurusAttribute)
                             the attribute to be added
           :param period: (float) polling period (in seconds)
           :param unsubscribe_evts: (bool) whether or not to unsubscribe from events
        """
        tmr = self.polling_timers.get(period, TaurusPollingTimer(period))
        self.polling_timers[period] = tmr
        tmr.addAttribute(attribute, self.isPollingEnabled())

    def removeAttributeFromPolling(self, attribute):
        """Deactivate the polling (client side) for the given attribute. If the
           polling of the attribute was not previously enabled, nothing happens.

           :param attribute: (taurus.core.taurusattribute.TaurusAttribute)
                             the attribute to be removed
        """
        timers = dict(self.polling_timers)
        for period, timer in timers.items():
            timer.removeAttribute(attribute)
            if not timer.dev_dict and period in self.polling_timers:
                del self.polling_timers[period]

    def __str__(self):
        return '{0}()'.format(self.__class__.__name__)

    def __repr__(self):
        return '{0}(schemes={1})'.format(self.__class__.__name__, ", ".join(self.schemes))

    def getValidTypesForName(self, name, strict=None):
        '''
        Returns a list of all Taurus element types for which `name` is a valid
        model name (while in many cases a name may only be valid for one
        element type, this is not necessarily true in general)

        In this base implementation, name is checked first for Attribute, then
        for Device and finally for Authority, and the return value is sorted in
        that same order.

        If a given schema requires a different ordering, reimplement this method

        :param name: (str) taurus model name

        :return: (list<TaurusElementType.element>) where element can be one of:
                 `Attribute`, `Device` or `Authority`
        '''
        ret = []
        if self.getAttributeNameValidator().isValid(name, strict=strict):
            ret.append(TaurusElementType.Attribute)
        if self.getDeviceNameValidator().isValid(name, strict=strict):
            ret.append(TaurusElementType.Device)
        if self.getAuthorityNameValidator().isValid(name, strict=strict):
            ret.append(TaurusElementType.Authority)
        return ret

    def getValidatorFromName(self, name):
        """
        Obtain the validator object corresponding to the given model
        name. If the model name is not valid for any TaurusModel class,
        it returns None
        """
        modeltypes = self.getValidTypesForName(name)
        if not modeltypes:
            return None
        return self.elementTypesMap[modeltypes[0]].getNameValidator()

    def findObjectClass(self, absolute_name):
        """
        Obtain the class object corresponding to the given name.

        Note, this generic implementation expects that derived classes provide a
        an attribute called elementTypesMap consisting in a dictionary whose
        keys are TaurusElementTypes and whose values are the corresponding
        specific object classes. e.g., the FooFactory should provide::

          class FooFactory(TaurusFactory):
              elementTypesMap = {TaurusElementType.Authority: FooAuthority,
                                 TaurusElementType.Device: FooDevice,
                                 TaurusElementType.Attribute: FooAttribute,
                                 }
              (...)


        :param absolute_name: (str) the object absolute name string

        :return: (taurus.core.taurusmodel.TaurusModel or None) a TaurusModel
                 class derived type or None if the name is not valid

        """
        try:
            elementTypesMap = self.elementTypesMap
        except AttributeError:
            msg = ('generic findObjectClass called but %s does ' +
                   'not define elementTypesMap.') % self.__class__.__name__
            raise RuntimeError(msg)
        for t in self.getValidTypesForName(absolute_name):
            ret = elementTypesMap.get(t, None)
            if ret is not None:
                return ret
        return None



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Unit tests for taurusbase"""


import pytest
from taurus.qt.qtgui.container import TaurusWidget

from taurus.test.pytest import check_taurus_deprecations

try:
    # The following are Tango-centric imports.
    from taurus.core.tango.test import taurus_test_ds  # pytest fixture

    _TANGO_MISSING = False
except:
    _TANGO_MISSING = True


@pytest.mark.skipif(_TANGO_MISSING, reason="tango-dependent test")
@pytest.mark.parametrize(
    "model, expected",
    [
        ("/boolean_scalar", "True"),
        ("/short_scalar", "123 mm"),
        ("/double_scalar", "1.23 mm"),
        ("/state", "ON"),
        ("/float_scalar#", "-----"),
        ("/float_scalar#label", "float_scalar"),
        ("/double_scalar#rvalue.magnitude", "1.23"),
        #("/float_scalar?configuration=label", "float_scalar"),
        ("eval:1+2", "3"),
        ("eval:1+2#label", "1+2"),
        ("eval:1+2#", "-----"),
    ],
)
def test_display_value(
    qtbot,
    caplog,
    taurus_test_ds,
    model,
    expected
):
    """Check the getDisplayValue method"""
    with check_taurus_deprecations(caplog):
        w = TaurusWidget()
        qtbot.addWidget(w)
        if model.startswith("/"):
            model = "tango:{}{}".format(taurus_test_ds, model)
        with qtbot.waitSignal(w.modelChanged, timeout=3200):
            w.setModel(model)

        def _ok():
            """Check text"""
            assert w.getDisplayValue() == expected

        qtbot.waitUntil(_ok, timeout=3200)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a set of basic taurus widgets based on QCheckBox"""

__all__ = ["TaurusValueCheckBox"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus.qt.qtgui.base import TaurusBaseWritableWidget


class TaurusValueCheckBox(Qt.QCheckBox, TaurusBaseWritableWidget):
    """A QCheckBox connected to a boolean writable attribute model"""

    def __init__(self, qt_parent=None, designMode=False):
        name = "TaurusValueCheckBox"
        self.call__init__wo_kw(Qt.QCheckBox, qt_parent)
        self.call__init__(TaurusBaseWritableWidget,
                          name, designMode=designMode)

        self.setObjectName(name)
        self.updateStyle()
        self.stateChanged.connect(self.notifyValueChanged)

    def keyPressEvent(self, event):
        if event.key() in (Qt.Qt.Key_Return, Qt.Qt.Key_Enter):
            self.writeValue()
            event.accept()
        else:
            Qt.QCheckBox.keyPressEvent(self, event)
            event.ignore()

    def minimumSizeHint(self):
        return Qt.QSize(20, 20)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWritableWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def updateStyle(self):
        TaurusBaseWritableWidget.updateStyle(self)
        # Show text only if it is not specifically hidden
        if self._showText:
            try:
                self.setText(str(self.getModelObj().getConfig().getLabel()))
            except:
                self.setText('----')
        else:
            self.setText('')
        # Update pending operations style
        if self.hasPendingOperations():
            txt = str(self.text()).strip()
            if len(txt) == 0:
                self.setText("!")
            self.setStyleSheet('TaurusValueCheckBox {color: blue;}')
        else:
            if str(self.text()) == "!":
                self.setText(" ")
            self.setStyleSheet('TaurusValueCheckBox {}')
        self.update()

    def setValue(self, v):
        self.setChecked(bool(v))

    def getValue(self):
        return self.isChecked()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWritableWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.input'
        ret['icon'] = "designer:checkbox.png"
        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", TaurusBaseWritableWidget.getModel,
                            TaurusBaseWritableWidget.setModel,
                            TaurusBaseWritableWidget.resetModel)

    showText = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getShowText,
                               TaurusBaseWritableWidget.setShowText,
                               TaurusBaseWritableWidget.resetShowText)

    useParentModel = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getUseParentModel,
                                     TaurusBaseWritableWidget.setUseParentModel,
                                     TaurusBaseWritableWidget.resetUseParentModel)

    autoApply = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getAutoApply,
                                TaurusBaseWritableWidget.setAutoApply,
                                TaurusBaseWritableWidget.resetAutoApply)

    forcedApply = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getForcedApply,
                                  TaurusBaseWritableWidget.setForcedApply,
                                  TaurusBaseWritableWidget.resetForcedApply)



import os
from taurus.external.qt.QtCore import PYSIDE2
import pytest


from taurus.qt.qtgui.graphic.jdraw import (
    TaurusJDrawGraphicsFactory,
    jdraw_parser
)
from taurus.qt.qtgui.graphic import TaurusGraphicsScene


@pytest.mark.skipif(PYSIDE2, reason="Avoid segfault when using PySide2")
def test_jdraw_parser(qtbot):
    """Check that jdraw_parser does not break with JDBar elements"""
    fname = os.path.join(os.path.dirname(__file__), "res", "bug1077.jdw")
    factory = TaurusJDrawGraphicsFactory(None)
    p = jdraw_parser.parse(fname, factory)
    assert isinstance(p, TaurusGraphicsScene)


if __name__ == '__main__':
    test_jdraw_parser()



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Unit tests for Taurus Label"""


from taurus.qt.qtgui.display import TaurusLabel
from taurus.core.util.colors import ATTRIBUTE_QUALITY_DATA, DEVICE_STATE_DATA
from taurus.external.qt import Qt
from pint import UnitRegistry as ur
import numpy

from taurus.test.pytest import check_taurus_deprecations

import pytest

try:
    # The following are Tango-centric imports.
    from taurus.core.tango.test import taurus_test_ds  # pytest fixture

    _TANGO_MISSING = False
except:
    _TANGO_MISSING = True


def _chek_tauruslabel(
    qtbot,
    caplog,
    taurus_test_ds,
    model,
    fmt=None,
    fgRole=None,
    bgRole=None,
    modelIndex=None,
    depr=0,
    expected_fg=None,
    expected_bg=None,
):
    """Check the label foreground and background"""
    # TODO: these tests are not properly isolated. For example, the
    #       parameterization testing fgrole="quality" fails in PySide2
    #       if it is called after another parameterization.
    if expected_fg is None and expected_bg is None:
        raise ValueError("expected_fg or expected_bg must not be None")
    with check_taurus_deprecations(caplog, expected=depr):
        w = TaurusLabel()
        qtbot.addWidget(w)
        if model.startswith("/"):
            model = "tango:{}{}".format(taurus_test_ds, model)
        with qtbot.waitSignal(w.modelChanged, timeout=3200):
            w.setModel(model)
        if fmt is not None:
            w.setFormat(fmt)
        if modelIndex is not None:
            w.setModelIndex(modelIndex)
        if fgRole is not None:
            w.setFgRole(fgRole)
        if bgRole is not None:
            w.setBgRole(bgRole)

        def _ok():
            """Check text"""
            if expected_fg is not None:
                assert w.text() == expected_fg
            if expected_bg is not None:
                p = w.palette()
                assert p.color(p.Background).getRgb()[:3] == expected_bg

        qtbot.waitUntil(_ok, timeout=3200)


@pytest.mark.skipif(_TANGO_MISSING, reason="tango-dependent test")
@pytest.mark.parametrize(
    "model, fgRole, modelIndex, depr, fg",
    [
        # pre-tep14 compat FgRole checks: value, w_value, state, quality, none
        ("/double_scalar", "quality", None, 0, "ATTR_VALID"),
        ("/double_scalar", "value", None, 0, "1.23 mm"),
        ("/double_scalar", "w_value", None, 0, "0.00 mm"),
        ("/double_scalar", "state", None, 0, "Ready"),
        ("/double_scalar", "none", None, 0, ""),
        # fragment and modelIndex checks
        ("/double_scalar#label", None, None, 0, "double_scalar"),
        ("/double_scalar", "label", None, 0, "double_scalar"),
        ("/double_scalar#rvalue", "label", None, 0, "double_scalar"),
        ("/double_scalar#state", None, None, 0, "Ready"),
        ("/double_spectrum", None, 1, 0, "1.23 mm"),
        ("/double_spectrum#rvalue[1]", None, None, 0, "1.23 mm"),
        ("/double_image", None, (1, 1), 0, "1.23 mm"),
        (
            "/double_image#rvalue[1,::2]",
            None,
            None,
            0,
            # expected is not explicit to support pint v<0.8 particularities
            "{:~}".format(numpy.array([1.23, 1.23]) * ur().mm),
        ),
    ],
)
def test_tauruslabel_text(
    qtbot, caplog, taurus_test_ds, model, fgRole, modelIndex, depr, fg
):
    """Check the label text"""
    _chek_tauruslabel(
        qtbot,
        caplog,
        taurus_test_ds,
        model,
        fgRole=fgRole,
        modelIndex=modelIndex,
        depr=depr,
        expected_fg=fg,
    )


TRANSPARENT_BG = Qt.QColor(Qt.Qt.transparent).getRgb()[:3]
TAURUS_READY_BG = DEVICE_STATE_DATA["TaurusDevState.Ready"][1:4]
TG_ATTR_VALID_BG = ATTRIBUTE_QUALITY_DATA["ATTR_VALID"][1:4]


@pytest.mark.skipif(_TANGO_MISSING, reason="tango-dependent test")
@pytest.mark.parametrize(
    "model, bgRole, bg",
    [
        # pre-tep14 compat BgRole checks: state, quality, none
        ("/float_scalar_ro", "none", TRANSPARENT_BG),
        ("/float_scalar_ro", "state", TAURUS_READY_BG),
        ("/float_scalar_ro", "quality", TG_ATTR_VALID_BG),
        ("/float_scalar_ro", None, TG_ATTR_VALID_BG),
    ],
)
def test_tauruslabel_bg(qtbot, caplog, taurus_test_ds, model, bgRole, bg):
    """"Check the label background"""
    _chek_tauruslabel(
        qtbot, caplog, taurus_test_ds, model, bgRole=bgRole, expected_bg=bg
    )


def _oneDecimalFormater(dtype, **kwargs):
    return "{:~.1f}"


def _typeFormatter(dtype, **kwargs):
    return dtype.__name__


@pytest.mark.parametrize(
    "model, fmt, fg",
    [
        ("eval:1.2345", "{:.3f}", "1.234 dimensionless"),
        ("eval:1.2345", "{:~.3f}", "1.234"),
        ("eval:1.2345", "{:~.3f}", "1.234"),
        ("eval:1.2345", ">>{}<<", ">>1.2345<<"),
        ('eval:"hello"', None, "hello"),
        ('eval:"hello"', _oneDecimalFormater, "hello"),
        ('eval:"hello"', _typeFormatter, "str"),
        ("eval:1.2345", _oneDecimalFormater, "1.2"),
        ('eval:Q("5m")#rvalue.units', _typeFormatter, "Unit"),
        ("eval:Q(5)#rvalue.magnitude", _typeFormatter, "int"),
    ],
)
def test_instance_format(qtbot, caplog, taurus_test_ds, model, fmt, fg):
    """Check formatter API at instance level"""
    _chek_tauruslabel(
        qtbot, caplog, taurus_test_ds, model, fmt=fmt, expected_fg=fg
    )


@pytest.mark.parametrize(
    "model, fmt, fg",
    [
        ("eval:1.2345", "{:.3f}", "1.234 dimensionless"),
        ("eval:1.2345", "{:~.3f}", "1.234"),
        ("eval:1.2345", "{:~.3f}", "1.234"),
        ("eval:1.2345", ">>{}<<", ">>1.2345<<"),
        ('eval:"hello"', None, "hello"),
        ('eval:"hello"', _oneDecimalFormater, "hello"),
        ('eval:"hello"', _typeFormatter, "str"),
        ("eval:1.2345", _oneDecimalFormater, "1.2"),
        ('eval:Q("5m")#rvalue.units', _typeFormatter, "Unit"),
        ("eval:Q(5)#rvalue.magnitude", _typeFormatter, "int"),
    ],
)
def test_class_format(monkeypatch, qtbot, caplog, model, fmt, fg):
    """Check formatter API at class level"""
    monkeypatch.setattr(TaurusLabel, "FORMAT", fmt)

    with check_taurus_deprecations(caplog, expected=0):
        w = TaurusLabel()
        qtbot.addWidget(w)

        w.setModel(model)
        w.resetFormat()  # needed to avoid fuzzyness in the tests

        def _ok():
            """Check text"""
            assert w.text() == fg

        qtbot.waitUntil(_ok, timeout=3200)


#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
.. currentmodule:: taurus.core.epics

Epics extension for taurus core model.

The epics extension provides access to Epics control system objects via Channel
Access

.. note:: The Epics scheme is only a proof of concept. The final syntax of
          the model names is not yet set in stone and only basic functionality
          is implemented.


The Epics extension implements :mod:`taurus.core` objects that connect to Epics
PVs. The scheme name for channel access epics models is 'ca' ('epics' also
works at this moment).

You should never instantiate models from epics model classes directly. Instead,
use the :class:`taurus.core.taurusmanager.TaurusManager` and
:class:`taurus.core.taurusmanager.TaurusFactory` APIs to access all elements.

For example, to get a reference to the epics process variable (PV)
"my:example.RBV" you should do something like::

    >>> import taurus
    >>> myattr = taurus.Attribute('ca:my:example.RBV')

Epics attributes (should) work just as other Taurus attributes and can be
referred by their model name wherever a Taurus Attribute model is expected. For
example, you can launch a `TaurusForm` with an epics attribute::

    $> taurus form ca:my:example

Similarly, you can combine epics attributes with attributes from other schemes::

    $> taurus form 'ca:my:example' 'tango:sys/tg_test/1/float_scalar'\
       'eval:{ca:my:example}*{tango:sys/tg_test/1/float_scalar}'

Currently, the taurus epics scheme just supports epics PVs, implementing them as
taurus attributes. Other model types such as the Authority, and Device classes
are just convenience dummy objects in the epics scheme at this point.
Epics records may eventually be mapped as Devices.
"""
from __future__ import absolute_import
from .epicsfactory import *



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Test for taurus.core.evaluation.test.test_evalfactory..."""

import re

import taurus
import unittest
from taurus.test import insertTest


@insertTest(helper_name='checkAttributeName', model='eval://1', oldstyle=True)
@insertTest(helper_name='checkAttributeName', model='a=2;a*3')
@insertTest(helper_name='checkAttributeName', model='1')
@insertTest(helper_name='checkAttributeID', model='eval://1', oldstyle=True)
@insertTest(helper_name='checkAttributeID', model='a=2;a*3')
@insertTest(helper_name='checkAttributeID', model='1')
@insertTest(helper_name='checkAttributeID', model='eval:1')
class EvaluationFactoryTestCase(unittest.TestCase):
    fragments = ['#', '#label', '#units']

    def setUp(self):
        self.f = taurus.Factory('eval')

    def convert2oldstyle(self, fragment):
        if fragment == '#':
            return '?configuration'
        return re.sub("#(?=.+)", "?configuration=", fragment)

    def checkAttributeID(self, model, oldstyle=False):
        '''Helper for test the attributes (by ID) when some different models
        of the same attribute are given (adding fragments in the models)
        '''
        attr = self.f.getAttribute(model)
        for fragment in self.fragments:
            if oldstyle:
                fragment = self.convert2oldstyle(fragment)
            attr2 = self.f.getAttribute(model + fragment)
            msg = '%s and %s has different id' % (attr.getFullName(),
                                                  attr2.getFullName())
            self.assertTrue(id(attr) == id(attr2), msg)

    def checkAttributeName(self, model, oldstyle=False):
        '''Helper for test the attribute names of the same attribute
        with different models (adding fragments in the models)
        '''
        attr = self.f.getAttribute(model)
        for fragment in self.fragments:
            if oldstyle:
                fragment = self.convert2oldstyle(fragment)
            attr2 = self.f.getAttribute(model + fragment)
            msg = '%s and %s has different ' % (attr.getFullName(),
                                                attr2.getFullName())
            self.assertTrue(attr.getFullName() == attr2.getFullName(),
                            msg + "fullname")
            self.assertTrue(attr.getNormalName() == attr2.getNormalName(),
                            msg + "normalname")
            self.assertTrue(attr.getSimpleName() == attr2.getSimpleName(),
                            msg + "simplename")



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains the graphics view widget for jdraw files"""
from __future__ import absolute_import

from builtins import str

import os
import traceback

from future.utils import string_types

import taurus
from taurus.external.qt import Qt, compat
from taurus.core.taurusbasetypes import TaurusElementType
from taurus.core import taurushelper
from taurus.qt.qtgui.graphic.taurusgraphic import parseTangoUri, TaurusGraphicsItem, SynopticSelectionStyle
from taurus.qt.qtcore.mimetypes import TAURUS_ATTR_MIME_TYPE, TAURUS_DEV_MIME_TYPE, TAURUS_MODEL_MIME_TYPE
from taurus.qt.qtgui.base import TaurusBaseWidget

from taurus.qt.qtgui.graphic.jdraw.jdraw_parser import parse

__all__ = ["TaurusJDrawSynopticsView"]

__docformat__ = 'restructuredtext'


class TaurusJDrawSynopticsView(Qt.QGraphicsView, TaurusBaseWidget):
    '''
    Taurus Class that visualizes Synoptics drawn with the JDraw tool (by ESRF). It is equivalent to ATK Synoptic Player (Java).

    After initialization call setModel('/your/file.jdw') to parse the synoptic file and connect to controlled objects.

    Arguments to TaurusJDrawSynopticsView() creator are:

        - designMode; used by Qt Designer
        - updateMode; controls Qt Viewport refresh (disabled by default)
        - alias; a dictionary of name replacements to be applied on graphical objects
        - resizable: whether to allow resizing or not
        - panelClass: class object, class name or shell command to be shown when an object is clicked (None will show default panel, '' or 'noPanel' will disable it)

    TaurusJDrawSynopticsView and TaurusGraphicsScene signals/slots

    External events::

     Slot selectGraphicItem(const QString &) displays a selection
     mark around the TaurusGraphicsItem that matches the argument passed.

    Mouse Left-button events::

     Signal graphicItemSelected(QString) is triggered, passing the
     selected TaurusGraphicsItem.name() as argument.

    Mouse Right-button events::

     TaurusGraphicsItem.setContextMenu([(ActionName,ActionMethod(device_name))]
     allows to configure custom context menus for graphic items using a list
     of tuples. Empty tuples will insert separators in the menu.
    '''
    itemsChanged = Qt.pyqtSignal('QString', dict)
    modelsChanged = Qt.pyqtSignal(list)
    graphicItemSelected = Qt.pyqtSignal('QString')
    graphicSceneClicked = Qt.pyqtSignal('QPoint')

    def __init__(self, parent=None, designMode=False, updateMode=None, alias=None, resizable=True, panelClass=None):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QGraphicsView, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)
        self._currF = self.modelName
        self.path = ''
        self.w_scene = None
        self.h_scene = None
        self._fileName = "Root"
        self._mousePos = (0, 0)
        self._selectionStyle = SynopticSelectionStyle.OUTLINE
        self.setResizable(resizable)
        self.setInteractive(True)
        self.setAlias(alias)
        self.setDragEnabled(True)
        self.setPanelClass(panelClass)

        # By default the items will update the view when necessary.
        # This default value is much more efficient then the QQraphicsView default
        # value, so if you decide to change then expect a lot of processor to be
        # used by your application.
        if updateMode is None:
            self.setViewportUpdateMode(Qt.QGraphicsView.NoViewportUpdate)
        else:
            self.setViewportUpdateMode(updateMode)

    def defineStyle(self):
        self.updateStyle()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget over writing
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def isReadOnly(self):
        return True

    def update(self):
        # self.emit_signal()
        self.emitColors()

    def openJDraw(self):
        ifile, _ = compat.getOpenFileName(
            self, 'Load JDraw File', '', 'JDraw File (*.jdw)')
        if not ifile:
            return
        fileName = ifile.split("/")
        self._fileName = fileName[-1]
        self.setModel(ifile)
        return fileName[-1]

    def setAlias(self, alias):
        """ Assigning a dictionary like {'Tag':'Value'} with tags to be replaced in object names while parsing. """
        if (isinstance(alias, dict) or hasattr(alias, 'items')) and alias:
            self.alias = alias
        else:
            self.alias = None
        return

    def get_item_list(self):
        return [item._name for item in self.scene().items() if hasattr(item, '_name') and item._name]

    def get_device_list(self):
        items = [(item, parseTangoUri(item)) for item in self.get_item_list()]
        return list(set(v['_devslashname'] for k, v in items if v))

    def get_item_colors(self, emit=False):
        item_colors = {}
        try:
            for item in self.scene().items():
                if not getattr(item, '_name', '') or not getattr(item, '_currBgBrush', None):
                    continue
                item_colors[item._name] = item._currBgBrush.color().name()
            if emit:
                self.itemsChanged.emit(self.modelName.split(
                    '/')[-1].split('.')[0], item_colors)
        except:
            self.warning('Unable to emitColors: %s' % traceback.format_exc())
        return item_colors

    @Qt.pyqtSlot(object)
    @Qt.pyqtSlot('QString')
    def selectGraphicItem(self, item_name):
        if self.scene() is not None:
            self.scene().selectGraphicItem(item_name)
        return False

    def _graphicItemSelected(self, item_name):
        self.debug(' => graphicItemSelected(QString)(%s)' % item_name)
        self.graphicItemSelected.emit(item_name)

    def _graphicSceneClicked(self, point):
        self.debug('In TaurusJDrawSynopticsView.graphicSceneClicked(%s,%s)' % (
            point.x(), point.y()))
        self.graphicSceneClicked.emit(point)

    def __modelsChanged(self):
        items = self.get_item_list()
        self.debug('modelsChanged(%s)' % len(items))
        self.modelsChanged.emit(items)

    def emitColors(self):
        '''emit signal which is used to refresh the tree and colors of icons depend of the current status in jdrawSynoptic'''
        self.get_item_colors(True)

    def get_sizes(self):
        srect = self.scene().sceneRect()
        sizes = [x for s in (self.size(), self.sizeHint(), srect.size())
                 for x in (s.width(), s.height())]
        try:
            s = self.parent().size()
            sizes.extend([s.width(), s.height()])
        except:
            sizes.extend([0, 0])
        return tuple(sizes)

    def fitting(self, ADJUST_FRAME=False):
        """
        Parent size is the size of the bigger panel (desn't keep ratio)
        Rect size never changes (fixed by the graphics objects)
        Size and SizeHint move one around the other

        the method works well until an object is clicked,
        then the whole reference changes and doesn't work again.
        """

        srect = self.scene().sceneRect()
        w, h = (srect.width(), srect.height())
        offset = self.mapToGlobal(Qt.QPoint(srect.x(), srect.y()))
        x0, y0 = (offset.x(), offset.y())

        self.debug('\n\nIn TauJDrawSynopticsView.fitting()')
        self.debug(self.get_sizes())
        self.debug(
            '\tAdjusting SizeHint: size(%s,%s),hint(%s,%s),srect(%s,%s),parent(%s,%s)' % self.get_sizes())
        self.debug('\toffset = %s,%s ; size = %s,%s' % (x0, y0, w, h))
        self.fitInView(x0, y0, w, h, Qt.Qt.KeepAspectRatio)

        # -------------------------------------------------------------------
        # This block seems to be a poorly executed workaround to some
        # issue, but it is itself causing bugs (see
        # https://github.com/taurus-org/taurus/issues/484 ).
        # We are not removing it yet to preserve bck-compat
        # TODO: Deprecate the ADJUST_FRAME kwarg?
        if ADJUST_FRAME:  # This additional resizing adjust the "white" frame around the synoptic
            self.debug(
                '\tResizing: size(%s,%s),hint(%s,%s),srect(%s,%s),parent(%s,%s)' % self.get_sizes())
            self.resize(self.sizeHint() + Qt.QSize(5, 5))
        # -------------------------------------------------------------------

        # THIS LINE MUST BE ALWAYS EXECUTED, It prevents the UP/DOWN resize BUG!!!
        # apparently Qt needs this 2 fitInView calls to be aware of it, maybe
        # first offset was not good
        self.debug(
            '\tFitting:: size(%s,%s),hint(%s,%s),srect(%s,%s),parent(%s,%s)' % self.get_sizes())
        self.fitInView(x0, y0, w, h, Qt.Qt.KeepAspectRatio)

        self.debug(
            'Done: size(%s,%s),hint(%s,%s),srect(%s,%s),parent(%s,%s)\n\n' % self.get_sizes())

    def resizeEvent(self, event):
        """ It has been needed to reimplent size policies """
        if not self.resizable() or not self.scene() or isinstance(self.parent(), Qt.QScrollArea) or not self.isVisible():
            self.debug('In TaurusJDrawSynopticsView(' +
                       self._fileName + ').resizeEvent(): Disabled')
            return
        try:
            self.debug('In TaurusJDrawSynopticsView(' +
                       self._fileName + ').resizeEvent()')
            if self.size() == self.sizeHint():
                self.debug('\tSize already fits: %s' % self.size())
                return
            self.setVerticalScrollBarPolicy(Qt.Qt.ScrollBarAlwaysOff)
            self.setHorizontalScrollBarPolicy(Qt.Qt.ScrollBarAlwaysOff)
            self.fitting()
            self.emitColors()
        except Exception:
            self.warning('Exception in JDrawView(' + self._fileName +
                         ').resizeEvent: %s' % traceback.format_exc())
            pass

    def refreshModel(self):
        self.setModel(self.getModelName())

    def updateStyle(self):
        self.repaint()

    def repaint(self):
        Qt.QGraphicsView.repaint(self)
        # self.fitting()

    def getGraphicsFactory(self, delayed=False):
        from . import jdraw
        # self.parent())
        return jdraw.TaurusJDrawGraphicsFactory(self, alias=(self.alias or None), delayed=delayed)

    ###########################################################################

    def getFramed(self):
        try:
            parent = self.parent()
        except:
            parent = None
        frame = Qt.QFrame(parent)
        self.setFrameStyle(self.StyledPanel | self.Raised)
        self.setLineWidth(2)
        self.setParent(frame)
        return frame

    def setResizable(self, resizable):
        self._resizable = resizable

    def resizable(self):
        return self._resizable

    def mousePressEvent(self, event):
        """ Records last event position to use it for DragEvents """
        try:
            self.mousePos = event.scenePos().x(), event.scenePos().y()
        except:
            self.mousePos = event.x(), event.y()
            self.debug('MouseEvent received is not a GraphicsScene event, using raw position %s' % str(
                self.mousePos))
        TaurusBaseWidget.mousePressEvent(self, event)

    def getModelMimeData(self):
        """ Used for drag events """
        model, mimeData = '', None
        try:
            #model = getattr(self.scene().itemAt(*self.mousePos),'_name','')
            selected = self.scene()._selectedItems
            if not selected:
                self.debug(
                    'jdrawView.getModelMimeData(%s): nothing to drag' % model)
                return
            model = getattr(([s for s in selected if s.isUnderMouse() and getattr(
                s, '_name', '')] or [selected])[0], '_name', '')
            self.debug('getModelMimeData(%s)' % model)
            mimeData = Qt.QMimeData()
            if model:
                taurusType = taurushelper.getValidTypesForName(model, False)
                model = str(model).encode('utf8')
                if TaurusElementType.Device in taurusType:
                    mimeData.setData(TAURUS_DEV_MIME_TYPE, model)
                if TaurusElementType.Attribute in taurusType:
                    mimeData.setData(TAURUS_ATTR_MIME_TYPE, model)
                else:
                    mimeData.setData(TAURUS_MODEL_MIME_TYPE, model)
        except:
            self.debug(
                'jdrawView.getModelMimeData(%s): unable to get MimeData' % model)
            self.debug(traceback.format_exc())
        return mimeData

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @classmethod
    def setDefaultPanelClass(klass, other):
        """
        This method returns the Class used to open new object panels on double-click (TaurusDevicePanel by default)
        """
        klass._defaultClass = other

    @classmethod
    def defaultPanelClass(klass):
        """
        This method assigns the Class used to open new object panels on double-click (TaurusDevicePanel by default)
        If an string is used it can be either a Taurus class or an OS launcher
        """
        if not hasattr(klass, '_defaultClass'):
            from taurus.qt.qtgui.panel import TaurusDevicePanel
            # 'taurusdevicepanel' #You can use an executable or a class
            klass._defaultClass = TaurusDevicePanel
        obj = klass._defaultClass
        return obj

    def setPanelClass(self, widget):
        self._panelClass = widget

    def panelClass(self):
        if self._panelClass is None:
            return self.defaultPanelClass()
        else:
            return self._panelClass

    @Qt.pyqtSlot('QString')
    def setModel(self, model, alias=None, delayed=False, trace=False):
        self.modelName = str(model)
        self._currF = str(model)
        if alias is not None:
            self.setAlias(alias)
        ll = taurus.getLogLevel()
        if trace:
            taurus.setLogLevel(taurus.Debug)
        # self.debug('setModel("%s")'%model)
        if self._currF:
            #filename = str(self._currFile.absoluteFilePath())
            filename = self._currF
            filename = os.path.realpath(filename)
            if os.path.isfile(filename):
                self.debug("Starting to parse %s" % filename)
                self.path = os.path.dirname(filename)
                factory = self.getGraphicsFactory(delayed=delayed)
                scene = parse(filename, factory)
                self.setSelectionStyle(self._selectionStyle)
                self.debug("Obtained %s(%s)", type(scene).__name__, filename)
                if not scene:
                    self.warning("TaurusJDrawSynopticsView.setModel(%s): Unable to parse %s!!!" % (
                        model, filename))
                elif self.w_scene is None and scene.sceneRect():
                    self.w_scene = scene.sceneRect().width()
                    self.h_scene = scene.sceneRect().height()
                else:
                    self.debug('JDrawView.sceneRect() is NONE!!!')
                self.setScene(scene)
                self.scene().graphicItemSelected.connect(self._graphicItemSelected)
                self.scene().graphicSceneClicked.connect(self._graphicSceneClicked)
                # Qt.QApplication.instance().lastWindowClosed.connect(self.close) #It caused a
                # segfault!
                self.__modelsChanged()
                self.setWindowTitle(self.modelName)
                # The emitted signal contains the filename and a dictionary
                # with the name of items and its color
                self.emitColors()  # get_item_colors(emit=True)
                self.fitting()
            else:
                self.setScene(None)
        #self.debug('out of setModel()')
        taurus.setLogLevel(ll)

    def closeEvent(self, event=None):
        if self.scene():
            self.scene().closeAllPanels()
        Qt.QGraphicsView.closeEvent(self, event)

    def setModels(self):
        """ This method triggers item.setModel(item._name) in all internal items. """
        for item in self.scene().items():
            if item._name and isinstance(item, TaurusGraphicsItem):
                self.debug(
                    'TaurusJDrawGraphicsFactory.setModels(): calling item.setModel(%s)' % (item._name))
                item.setModel(item._name)

    def getModel(self):
        return self._currF

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['group'] = 'Taurus Display'
        ret['module'] = 'taurus.qt.qtgui.graphic'
        ret['icon'] = "designer:graphicsview.png"
        return ret

    model = Qt.pyqtProperty("QString", getModel, setModel)

    def setSelectionStyle(self, selectionStyle):
        if isinstance(selectionStyle,  string_types):
            selectionStyle = str(selectionStyle).upper()
            try:
                selectionStyle = SynopticSelectionStyle[selectionStyle]
            except:
                self.debug('invalid selectionStyle "%s"', selectionStyle)
                return
        if self.scene() is not None:
            self.setSelectionStyle(selectionStyle)
        self._selectionStyle = selectionStyle

    def getSelectionStyle(self):
        return self._selectionStyle

    def getSelectionStyleName(self):
        return SynopticSelectionStyle.whatis(self.getSelectionStyle())

    def resetSelectionStyle(self):
        self.setSelectionStyle(SynopticSelectionStyle.OUTLINE)

    selectionStyle = Qt.pyqtProperty("QString", getSelectionStyleName,
                                     setSelectionStyle,
                                     resetSelectionStyle)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


def jdraw_view_main():
    import sys
    import time
    import taurus.qt.qtgui.graphic
    taurus.setLogLevel(taurus.Info)
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv, cmd_line_parser=None)

    #form = Qt.QDialog()
    # ly=Qt.QVBoxLayout(form)
    # container=Qt.QWidget()
    # ly.addWidget(container)
    # for m in sys.argv[1:]:
    #tv=TaurusJDrawSynopticsView(container, designMode=False)
    # tv.setModel(m)
    form = taurus.qt.qtgui.graphic.TaurusJDrawSynopticsView(designMode=False)
    form.show()
    # print '%s setModel(%s)'%(time.ctime(),sys.argv[1])
    form.setModel(sys.argv[1])
    form.setWindowTitle(sys.argv[1].rsplit('.', 1)[0])
    sys.exit(app.exec_())

if __name__ == "__main__":
    jdraw_view_main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains all taurus tango authority"""

from __future__ import print_function

from builtins import str
from builtins import map
from builtins import range
from builtins import object
try:
    from collections.abc import Mapping
except ImportError:  # bck-compat py 2.7
    from collections import Mapping

import os
import weakref

from PyTango import (Database, DeviceProxy, DevFailed, ApiUtil)
from taurus import Device
from taurus.core.taurusbasetypes import TaurusDevState, TaurusEventType
from taurus.core.taurusauthority import TaurusAuthority
from taurus.core.util.containers import CaselessDict
from taurus.core.util.log import taurus4_deprecation
from taurus.core.util.fqdn import fqdn_no_alias


__all__ = ["TangoInfo", "TangoAttrInfo", "TangoDevInfo", "TangoServInfo",
           "TangoDevClassInfo", "TangoDatabaseCache", "TangoDatabase",
           "TangoAuthority"]

__docformat__ = "restructuredtext"

InvalidAlias = "nada"


class TangoInfo(object):

    def __init__(self, container, name=None, full_name=None):
        self._container = weakref.ref(container)
        self._name = name
        self._full_name = full_name

    def container(self):
        return self._container()

    def name(self):
        return self._name

    def fullName(self):
        return self._full_name

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.name())
    __repr__ = __str__


class TangoAttrInfo(TangoInfo):

    def __init__(self, container, name=None, full_name=None, device=None,
                 info=None):
        super(TangoAttrInfo, self).__init__(container, name=name,
                                            full_name=full_name)
        self._info = info
        self._device = weakref.ref(device)

    def device(self):
        return self._device()

    def info(self):
        return self._info

    def __getattr__(self, name):
        return getattr(self._info, name)


class TangoDevClassInfo(TangoInfo):

    def __init__(self, container, name=None, full_name=None):
        super(TangoDevClassInfo, self).__init__(container, name=name,
                                                full_name=full_name)

        self._devices = CaselessDict()

    def devices(self):
        return self._devices

    def addDevice(self, dev):
        self._devices[dev.name()] = dev

    def getDeviceNames(self):
        if not hasattr(self, "_device_name_list"):
            self._device_name_list = sorted(map(TangoDevInfo.name,
                                                self._devices.values()))
        return self._device_name_list


class TangoDevInfo(TangoInfo):

    def __init__(self, container, name=None, full_name=None, alias=None,
                 server=None, klass=None, exported=False, host=None):
        super(TangoDevInfo, self).__init__(container, name=name,
                                           full_name=full_name)
        self._alias = alias
        self._server = weakref.ref(server)
        self._klass = weakref.ref(klass)
        self._exported = bool(int(exported))
        self._alive = None
        self._state = None
        self._host = host
        name = str(name) # python2 compatibility
        self._domain, self._family, self._member = list(map(str.upper,
                                                       name.split("/", 2)))
        self._attributes = None
        self._alivePending = False

    def domain(self):
        return self._domain

    def family(self):
        return self._family

    def member(self):
        return self._member

    def alias(self):
        return self._alias

    def server(self):
        return self._server()

    def klass(self):
        return self._klass()

    def exported(self):
        return self._exported

    def alive(self):
        if self._alive is None:
            if self._alivePending:
                return False
            self._alivePending = True
            try:
                dev = self.getDeviceProxy()
                _ = dev.state()
                self._alive = True
            except:
                self._alive = False
            self._alivePending = False
        return self._alive

    def state(self):
        """Overwrite state so it doesn't call 'alive()' since it can take
        a long time for devices that are declared as exported but are in fact
        not running (crashed, network error, power cut, etc)"""
        if not self._state is None:
            return self._state
        exported = self.exported()
        if exported:
            self._state = TaurusDevState.Ready
        else:
            self._state = TaurusDevState.NotReady
        return self._state

    def host(self):
        return self._host

    def attributes(self):
        if self._attributes is None or len(self._attributes) == 0:
            self.refreshAttributes()
        return self._attributes

    def getAttribute(self, attrname):
        attrname = attrname.lower()
        for a in self.attributes():
            if a.name() == attrname:
                return a
        return None

    def setAttributes(self, attributes):
        self._attributes = attributes

    @taurus4_deprecation(alt="getDeviceProxy()")
    def getHWObj(self):
        return self.getDeviceProxy()

    def getDeviceProxy(self):
        db = self.container().db
        name = self.name()
        full_name = db.getFullName() + "/" + name
        dev = None
        try:
            dev = db.factory().getDevice(full_name).getDeviceProxy()
        except:
            pass
        return dev

    def refreshAttributes(self):
        attrs = []
        try:
            dev = self.getDeviceProxy()
            if dev is None:
                raise DevFailed()  # @todo: check if this is the right exception to throw
            attr_info_list = dev.attribute_list_query_ex()
            for attr_info in attr_info_list:
                full_name = "%s/%s" % (self.fullName(), attr_info.name)
                attr_obj = TangoAttrInfo(self.container(),
                                         name=attr_info.name.lower(), full_name=full_name.lower(),
                                         device=self, info=attr_info)
                attrs.append(attr_obj)
            attrs = sorted(attrs, key=lambda attr: attr.name())
        except DevFailed as df:
            self._state = TaurusDevState.NotReady
        self.setAttributes(attrs)


class TangoServInfo(TangoInfo):

    def __init__(self, container, name=None, full_name=None):
        super(TangoServInfo, self).__init__(container, name=name,
                                            full_name=full_name)
        self._devices = {}
        self._exported = False
        self._alive = None
        self._host = ""
        self._server_name, self._server_instance = name.split("/", 1)
        self._alivePending = False

    def devices(self):
        return self._devices

    def getDeviceNames(self):
        if not hasattr(self, "_device_name_list"):
            self._device_name_list = sorted(map(TangoDevInfo.name, self._devices.values()))
        return self._device_name_list

    def getClassNames(self):
        if not hasattr(self, "_klass_name_list"):
            klasses = set(map(TangoDevInfo.klass, self._devices.values()))
            self._klass_name_list = sorted(map(TangoDevClassInfo.name,
                                               klasses))
        return self._klass_name_list

    def exported(self):
        return self._exported

    def state(self):
        exported = self.exported()
        if exported:
            return TaurusDevState.Ready
        return TaurusDevState.NotReady

    def host(self):
        return self._host

    def serverName(self):
        return self._server_name

    def serverInstance(self):
        return self._server_instance

    def addDevice(self, dev):
        self._exported |= dev.exported()
        self._host = dev.host()
        self._devices[dev.name()] = dev

    def alive(self):
        if self._alive is None:
            if self._alivePending:
                return False
            try:
                self._alivePending = True
                alive = True
                for d in self.devices().values():
                    alive = d.alive()
                    if not alive:
                        break
                self._alive = alive
            except Exception as e:
                print("except", e)
                self._alive = False
            self._alivePending = False
        return self._alive


class TangoDatabaseCache(object):

    def __init__(self, db):
        self._db = weakref.ref(db)
        self._device_tree = None
        self._server_tree = None
        self._servers = None
        self._server_name_list = None
        self._devices = None
        self._device_name_list = None
        self._klasses = None
        self._klass_name_list = None
        self._aliases = None
        self._alias_name_list = None
        self.refresh()

    @property
    def db(self):
        return self._db()

    def refresh(self):
        db = self.db
        db_dev_name = '/'.join((db.getFullName(), db.dev_name()))
        if hasattr(Device(db_dev_name), 'DbMySqlSelect'):
            # optimization in case the db exposes a MySQL select API
            query = ("SELECT name, alias, exported, host, server, class " +
                     "FROM device")
            r = db.command_inout("DbMySqlSelect", query)
            row_nb, column_nb = r[0][-2:]
            data = r[1]
            assert row_nb == len(data) // column_nb
        else:
            # fallback using tango commands (slow but works with sqlite DB)
            # see http://sf.net/p/tauruslib/tickets/148/
            data = []
            all_alias = {}
            all_devs = db.get_device_name('*', '*')
            all_exported = db.get_device_exported('*')
            for k in db.get_device_alias_list('*'):  # Time intensive!!
                all_alias[db.get_device_alias(k)] = k
            for d in all_devs:  # Very time intensive!!
                _info = db.command_inout("DbGetDeviceInfo", d)[1]
                name, ior, level, server, host, started, stopped = _info[:7]
                klass = db.get_class_for_device(d)
                alias = all_alias.get(d, '')
                exported = str(int(d in all_exported))
                data.extend((name, alias, exported, host, server, klass))
            column_nb = 6  # len ((name, alias, exported, host, server, klass))

        CD = CaselessDict
        dev_dict, serv_dict, klass_dict, alias_dict = CD(), {}, {}, CD()

        for i in range(0, len(data), column_nb):
            name, alias, exported, host, server, klass = data[i:i + column_nb]
            if name.count("/") != 2:
                continue  # invalid/corrupted entry: just ignore it
            if server.count("/") != 1:
                continue  # invalid/corrupted entry: just ignore it
            if not len(alias):
                alias = None

            serv_dict[server] = si = serv_dict.get(server,
                                                   TangoServInfo(self, name=server,
                                                                 full_name=server))

            klass_dict[klass] = dc = klass_dict.get(klass,
                                                    TangoDevClassInfo(self,
                                                                      name=klass,
                                                                      full_name=klass))

            full_name = "%s/%s" % (db.getFullName(), name)
            dev_dict[name] = di = TangoDevInfo(self, name=name, full_name=full_name,
                                               alias=alias, server=si, klass=dc,
                                               exported=exported, host=host)

            si.addDevice(di)
            dc.addDevice(di)
            if alias is not None:
                alias_dict[alias] = di

        self._devices = dev_dict
        self._device_tree = TangoDevTree(dev_dict)
        self._server_tree = TangoServerTree(serv_dict)
        self._servers = serv_dict
        self._klasses = klass_dict
        self._aliases = alias_dict

    def refreshAttributes(self, device):
        attrs = []
        try:
            db = self.db
            name = device.name()
            full_name = db.getFullName() + "/" + name
            taurus_dev = db.factory().getDevice(full_name,
                                                create_if_needed=False)
            if taurus_dev is None:
                dev = DeviceProxy(full_name)
            else:
                dev = taurus_dev.getDeviceProxy()
            attr_info_list = dev.attribute_list_query_ex()
            for attr_info in attr_info_list:
                full_attr_name = "%s/%s" % (full_name, attr_info.name)
                attr_obj = TangoAttrInfo(self, name=attr_info.name,
                                         full_name=full_attr_name,
                                         device=device, info=attr_info)
                attrs.append(attr_obj)
            attrs = sorted(attrs, key=lambda attr: attr.name().lower())
        except DevFailed as df:
            pass
        device.setAttributes(attrs)

    def getDevice(self, name):
        """Returns a :class:`TangoDevInfo` object with information
        about the given device name

        :param name: (str) the device name

        :return: (TangoDevInfo) information about the device"""
        return self._devices.get(name)

    def getDeviceNames(self):
        """Returns a list of registered device names

        :return: (sequence<str>) a sequence with all registered device names"""
        if self._device_name_list is None:
            self._device_name_list = sorted(
                map(TangoDevInfo.name, self.devices().values()))
        return self._device_name_list

    def getAliasNames(self):
        if self._alias_name_list is None:
            self._alias_name_list = sorted(
                map(TangoDevInfo.alias, self.aliases().values()))
        return self._alias_name_list

    def getServerNames(self):
        """Returns a list of registered server names

        :return: (sequence<str>) a sequence with all registered server names"""
        if self._server_name_list is None:
            self._server_name_list = sorted(
                map(TangoServInfo.name, self.servers().values()))
        return self._server_name_list

    def getClassNames(self):
        """Returns a list of registered device classes

        :return: (sequence<str>) a sequence with all registered device classes"""
        if self._klass_name_list is None:
            self._klass_name_list = sorted(
                map(TangoDevClassInfo.name, self.klasses().values()))
        return self._klass_name_list

    def deviceTree(self):
        """Returns a tree container with all devices in three levels: domain,
           family and member

           :return: (TangoDevTree) a tree containning all devices"""
        return self._device_tree

    def serverTree(self):
        """Returns a tree container with all servers in two levels: server name
        and server instance

           :return: (TangoServerTree) a tree containning all servers"""
        return self._server_tree

    def servers(self):
        return self._servers

    def devices(self):
        return self._devices

    def klasses(self):
        return self._klasses

    def getDeviceDomainNames(self):
        return list(self._device_tree.keys())

    def getDeviceFamilyNames(self, domain):
        families = self._device_tree.get(domain)
        if families is None:
            return []
        return list(families.keys())

    def getDeviceMemberNames(self, domain, family):
        families = self._device_tree.get(domain)
        if families is None:
            return []
        members = families.get(family)
        if members is None:
            return []
        return list(members.keys())

    def getDomainDevices(self, domain):
        return self.deviceTree().getDomainDevices(domain)

    def getFamilyDevices(self, domain, family):
        return self.deviceTree().getFamilyDevices(domain, family)

    def getServerNameInstances(self, serverName):
        return self.serverTree().getServerNameInstances(serverName)


class TangoDevTree(CaselessDict):

    def __init__(self, other=None):
        super(TangoDevTree, self).__init__()
        self._devices = CaselessDict()
        if other is not None:
            self._update(other)

    def _update(self, other):
        try:
            if isinstance(other, Mapping):
                other = list(other.values())
            for dev in other:
                try:
                    self.addDevice(dev)
                except Exception as e:
                    print(e)
        except Exception as e:
            raise Exception(
                "Must give dict<obj, TangoDevInfo> or sequence<TangoDevInfo>")

    def addDevice(self, dev_info):
        domain, family, member = dev_info.domain(), dev_info.family(), dev_info.member()

        families = self[domain] = self.get(domain, CaselessDict())
        devs = self._devices[domain] = self._devices.get(
            domain, CaselessDict())
        devs[dev_info.name()] = dev_info

        families[family] = members = families.get(family, CaselessDict())

        members[member] = dev_info

    def getDomainDevices(self, domain):
        """Returns all devices under the given domain. Returns empty list if
        the domain doesn't exist or doesn't contain any devices"""
        return list(self._devices.get(domain, {}).values())

    def getFamilyDevices(self, domain, family):
        """Returns all devices under the given domain/family. Returns empty list if
        the domain/family doesn't exist or doesn't contain any devices"""
        families = self.get(domain)
        if families is None:
            return
        return list(families.get(family, {}).values())


class TangoServerTree(dict):

    def __init__(self, other=None):
        super(TangoServerTree, self).__init__()
        if other is not None:
            self._update(other)

    def _update(self, other):
        try:
            if isinstance(other, Mapping):
                other = list(other.values())
            for serv in other:
                try:
                    self.addServer(serv)
                except Exception as e:
                    print(e)
        except Exception as e:
            raise Exception(
                "Must give dict<obj, TangoServInfo> or sequence<TangoServInfo>")

    def addServer(self, serv_info):
        serverName, serverInstance = serv_info.serverName(), serv_info.serverInstance()

        serverInstances = self[serverName] = self.get(serverName, {})

        serverInstances[serverInstance] = serv_info

    def getServerNameInstances(self, serverName):
        """Returns all servers under the given serverName. Returns empty list if
        the server name doesn't exist or doesn't contain any instances"""
        return list(self.get(serverName, {}).values())


def get_home():
    """
    Find user's home directory if possible. Otherwise raise error.

    :return: user's home directory
    :rtype: str

    New in PyTango 7.2.0
    """
    path = ''
    try:
        path = os.path.expanduser("~")
    except:
        pass
    if not os.path.isdir(path):
        for evar in ('HOME', 'USERPROFILE', 'TMP'):
            try:
                path = os.environ[evar]
                if os.path.isdir(path):
                    break
            except:
                pass
    if path:
        return path
    else:
        raise RuntimeError('please define environment variable $HOME')


def get_env_var(env_var_name):
    """
    Returns the value for the given environment name
    A backup method for old Tango/PyTango versions which don't implement
    :meth:`PyTango.ApiUtil.get_env_var`

    Search order:

        * a real environ var
        * HOME/.tangorc
        * /etc/tangorc

    :param env_var_name: the environment variable name
    :type env_var_name: str
    :return: the value for the given environment name
    :rtype: str
    """

    if env_var_name in os.environ:
        return os.environ[env_var_name]

    fname = os.path.join(get_home(), '.tangorc')
    if not os.path.exists(fname):
        if os.name == 'posix':
            fname = "/etc/tangorc"
    if not os.path.exists(fname):
        return None

    with open(fname) as f:
        for line in f:
            strippedline = line.split('#', 1)[0].strip()

            if not strippedline:
                # empty line
                continue

            tup = strippedline.split('=', 1)
            if len(tup) != 2:
                # illegal line!
                continue

            key, val = list(map(str.strip, tup))
            if key == env_var_name:
                return val


class TangoAuthority(TaurusAuthority):

    # helper class property that stores a reference to the corresponding
    # factory
    _factory = None
    _scheme = 'tango'
    _description = 'A Tango Authority'

    def __init__(self, host=None, port=None, parent=None):
        if host is None or port is None:
            try:
                _hp = TangoAuthority.get_default_tango_host()
                host, port = _hp.rsplit(':', 1)
            except Exception:
                from taurus import warning
                warning("Error getting default Tango host")

        # Set host to fqdn
        host = fqdn_no_alias(host)

        self.dbObj = Database(host, port)
        self._dbProxy = None
        self._dbCache = None

        complete_name = "tango://%s:%s" % (host, port)
        self.call__init__(TaurusAuthority, complete_name, parent)

        try:
            self.get_class_for_device(self.dev_name())
        except:
            # Ok, old tango database.
            self.get_class_for_device = self.__get_class_for_device

    @staticmethod
    def get_default_tango_host():
        if hasattr(ApiUtil, "get_env_var"):
            f = ApiUtil.get_env_var
        else:
            f = get_env_var
        return f("TANGO_HOST")

    def __get_class_for_device(self, dev_name):
        """Backup method when connecting to tango 5 database device server"""
        # Ok, old tango database.
        serv_name = self.command_inout("DbGetDeviceInfo", dev_name)[1][3]
        devs = self.get_device_class_list(serv_name)
        dev_name_lower = dev_name.lower()
        for i in range(len(devs) // 2):
            idx = i * 2
            if devs[idx].lower() == dev_name_lower:
                return devs[idx + 1]
        return None

    def get_device_attribute_list(self, dev_name, wildcard):
        return self.command_inout("DbGetDeviceAttributeList", (dev_name, wildcard))

    # Export the PyTango.Database interface into this object.
    # This way we can call for example get_attribute_property on an object of
    # this class
    def __getattr__(self, name):
        if not self.dbObj is None:
            return getattr(self.dbObj, name)
        return None

    @taurus4_deprecation(alt='getTangoDB')
    def getValueObj(self, cache=True):
        return self.getTangoDB()

    def getTangoDB(self):
        return self.dbObj

    @taurus4_deprecation(alt='getFullName')
    def getDisplayValue(self, cache=True):
        return self.getDisplayDescription(cache)

    def addListener(self, listener):
        ret = TaurusAuthority.addListener(self, listener)
        if not ret:
            return ret
        self.fireEvent(TaurusEventType.Change, self.getFullName(), listener)
        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Query capabilities built on top of a cache
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def cache(self):
        if self._dbCache is None:
            self._dbCache = TangoDatabaseCache(self)
        return self._dbCache

    def refreshCache(self):
        self.cache().refresh()

    def getDevice(self, name):
        """
        Reimplemented from :class:`TaurusDevice` to use cache and return
        :class:`taurus.core.tango.TangoDevInfo` objects with information
        about the given device name

        :param name: (str) the device name

        :return: (TangoDevInfo) information about the tango device"""
        return self.cache().getDevice(name)

    def getDeviceNames(self):
        """Returns a list of registered tango device names

        :return: (sequence<str>) a sequence with all registered tango device names"""
        return self.cache().getDeviceNames()

    def getAliasNames(self):
        """Returns a list of registered tango device alias

        :return: (sequence<str>) a sequence with all registered tango device alias"""
        return self.cache().getAliasNames()

    def getServerNames(self):
        """Returns a list of registered tango device servers in format<name>/<instance>

        :return: (sequence<str>) a sequence with all registered tango device servers"""
        return self.cache().getServerNames()

    def getClassNames(self):
        """Returns a list of registered tango device classes

        :return: (sequence<str>) a sequence with all registered tango device classes"""
        return self.cache().getClassNames()

    def getDeviceDomainNames(self):
        return self.cache().getDeviceDomainNames()

    def getDeviceFamilyNames(self, domain):
        return self.cache().getDeviceFamilyNames(domain)

    def getDeviceMemberNames(self, domain, family):
        return self.cache().getDeviceMemberNames(domain, family)

    def getDomainDevices(self, domain):
        return self.cache().getDomainDevices(domain)

    def getFamilyDevices(self, domain, family):
        return self.cache().getFamilyDevices(domain, family)

    def getServerNameInstances(self, serverName):
        return self.cache().getServerNameInstances(serverName)

    def deviceTree(self):
        """Returns a tree container with all devices in three levels : domain,
           family and member

           :return: (TangoDevTree) a tree containning all devices"""
        return self.cache().deviceTree()

    def getElementAlias(self, full_name):
        '''return the alias of an element from its full name'''
        try:
            alias = self.getTangoDB().get_alias(full_name)
            if alias and alias.lower() == InvalidAlias:
                alias = None
        except:
            alias = None
        return alias

    def getElementFullName(self, alias):
        '''return the full name of an element from its alias'''
        try:  # PyTango v>=8.1.0
            return self.getTangoDB().get_device_from_alias(alias)
        except AttributeError:
            try:  # PyTango v<8.1.0
                return self.getTangoDB().get_device_alias(alias)
            except:
                return None
        except:
            return None

    @taurus4_deprecation(alt=".description")
    def getDescription(self, cache=True):
        return self.description


# Declare this alias for backwards compatibility
TangoDatabase = TangoAuthority



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains taurus Qt form widgets"""

from __future__ import print_function
from __future__ import absolute_import

import click
from datetime import datetime
from functools import partial

from future.utils import string_types, binary_type

from taurus import tauruscustomsettings as _ts
from taurus.external.qt import Qt

import taurus.core
from taurus.core import DisplayLevel

from taurus.qt.qtcore.mimetypes import (TAURUS_ATTR_MIME_TYPE, TAURUS_DEV_MIME_TYPE,
                                        TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_MODEL_MIME_TYPE)
from taurus.qt.qtgui.container import TaurusWidget, TaurusScrollArea
from taurus.qt.qtgui.button import QButtonBox, TaurusCommandButton
from .taurusmodelchooser import TaurusModelChooser
from taurus.core.util.log import deprecation_decorator
from taurus.core.util.plugin import selectEntryPoints
from taurus import warning
import taurus.cli.common


__all__ = ["TaurusAttrForm", "TaurusCommandsForm", "TaurusForm"]

__docformat__ = 'restructuredtext'


def _normalize_model_name_case(modelname):
    """
    Accepts a model name and returns it in lower case if the models is
    case insensitive. Otherwise it returns the same model name
    """
    if taurus.Factory(taurus.getSchemeFromName(modelname)).caseSensitive:
        return modelname
    else:
        return modelname.lower()


class ParameterCB(Qt.QComboBox):
    '''A custom combobox'''

    def __init__(self, parent=None):
        Qt.QComboBox.__init__(self, parent)
        self.setEditable(True)

    def rememberCurrentText(self):
        '''Adds the current text to the combobox items list (unless it is already there)'''
        text = self.currentText()
        if self.findText(text) < 0:
            self.addItem(text)


class TaurusForm(TaurusWidget):
    '''A form containing specific widgets for interacting with
    a given list of taurus attributes and/or devices.

    Its model is a list of attribute and/or device names to be shown. Each item
    is represented in a row consisting of a label, a read widget, a write
    widget, a units widget and an "extra" widget (some of them may not be shown)
    which are vertically aligned with their counterparts from other items.

    By default a :class:`TaurusValue` object is used for each item, but this
    can be changed and customizations can be provided by enabling/disabling
    item factories with :meth:`setItemFactories`.

    Item objects can be accessed by index using a list-like notation::

      form = TaurusForm()
      form.model = ['sys/tg_test/1'+a for a in ('short_image','/float_scalar','/double_scalar')]
      form[0].labelConfig = 'dev_alias'
      form[-1].writeWidgetClass = 'TaurusWheelEdit'
      print(len(form))  # --> outputs '3' (the length of the form is the number of items)

    By default, the form provides global Apply and Cancel buttons.

    You can also see some code that exemplifies the use of TaurusForm in :ref:`Taurus
    coding examples <examples>` '''

    def __init__(self, parent=None,
                 formWidget=None,  # deprecated
                 buttons=None,
                 withButtons=True,
                 designMode=False):

        self._children = []
        TaurusWidget.__init__(self, parent, designMode)

        if buttons is None:
            buttons = Qt.QDialogButtonBox.Apply | \
                Qt.QDialogButtonBox.Reset
        self._customWidgetMap = {}  # deprecated
        self._model = []

        if formWidget is None:
            from taurus.qt.qtgui.panel import TaurusValue
            formWidget = TaurusValue
        else:
            self.deprecated(
                dep="formWidget argument", alt="item factories", rel="4.6.5")
        self._defaultFormWidget = formWidget

        self._itemFactories = []
        self.setItemFactories()

        self.setLayout(Qt.QVBoxLayout())

        frame = TaurusWidget()
        frame.setLayout(Qt.QGridLayout())

        self.scrollArea = TaurusScrollArea(self)
        self.scrollArea.setWidget(frame)
        self.scrollArea.setWidgetResizable(True)
        self.layout().addWidget(self.scrollArea)
        self.__modelChooserDlg = None

        self.buttonBox = QButtonBox(buttons=buttons, parent=self)
        self.layout().addWidget(self.buttonBox)

        self._connectButtons()

        # Actions (they automatically populate the context menu)
        self.setContextMenuPolicy(Qt.Qt.ActionsContextMenu)

        self.chooseModelsAction = Qt.QAction('Modify Contents', self)
        self.addAction(self.chooseModelsAction)
        self.chooseModelsAction.triggered.connect(self.chooseModels)

        self.showButtonsAction = Qt.QAction('Show Buttons', self)
        self.showButtonsAction.setCheckable(True)
        self.addAction(self.showButtonsAction)
        self.showButtonsAction.triggered[bool].connect(self.setWithButtons)
        self.setWithButtons(withButtons)

        self.changeLabelsAction = Qt.QAction('Change labels (all items)', self)
        self.addAction(self.changeLabelsAction)
        self.changeLabelsAction.triggered.connect(self.onChangeLabelsAction)

        self.compactModeAction = Qt.QAction('Compact mode (all items)', self)
        self.compactModeAction.setCheckable(True)
        self.addAction(self.compactModeAction)
        self.compactModeAction.triggered[bool].connect(self.setCompact)

        self.setFormatterAction = Qt.QAction('Set formatter (all items)', self)
        self.addAction(self.setFormatterAction)
        self.setFormatterAction.triggered.connect(self.onSetFormatter)

        self.resetModifiableByUser()
        self.setSupportedMimeTypes([TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_DEV_MIME_TYPE,
                                    TAURUS_ATTR_MIME_TYPE, TAURUS_MODEL_MIME_TYPE, 'text/plain'])

        self.resetCompact()

        # properties
        self.registerConfigProperty(
            self.isWithButtons, self.setWithButtons, 'withButtons')
        self.registerConfigProperty(self.isCompact, self.setCompact, 'compact')

    def __getitem__(self, key):
        '''provides a list-like interface: items of the form can be accessed using slice notation'''
        return self.getItemByIndex(key)

    def __len__(self):
        '''returns the number of items contained by the form'''
        return len(self.getItems())

    def setItemFactories(self, include=None, exclude=None):
        """
        Selects and prioritizes the factories to be used to create the form's
        items.

        TaurusForm item factories are functions that receive a TaurusModel as
        their only argument and return either a TaurusValue-like instance
        or None in case the factory does not handle the given model.

        The factories are selected using their entry point names as registered
        in the "taurus.form.item_factories" entry point group.

        The factories entry point name is up to the registrar of the entry
        point (typically a taurus plugin) and should be documented by the
        registrar to allow for selection and prioritization.

        The selection and prioritization is done using
        :meth:`taurus.core.util.plugin.selectEntryPoints()`. See it for
        more details.

        The selected list is updated in the form, and returned.

        The default values for the include and exclude arguments are defined
        in `tauruscustomsettings.T_FORM_ITEM_FACTORIES`

        :param include: (tuple). The members in the tuple can be: Regexp
                        patterns (in string or compiled form) matching the
                        names to be included in the selection. They can also be
                        item factory functions (which then are wrapped in an
                        EntryPoint-like object and included in the selection).
        :param exclude: (tuple). Regexp patterns ( either `str` or
                        :class:`re.Pattern` objects) matching registered names
                        to be excluded.
        :return: (list) selected item factories entry points
        """
        patterns = getattr(_ts, "T_FORM_ITEM_FACTORIES", {})
        if include is None:
            include = patterns.get("include", ('.*',))
        if exclude is None:
            exclude = patterns.get("exclude", ())
        self._itemFactories = selectEntryPoints(
            group='taurus.form.item_factories',
            include=include,
            exclude=exclude
        )
        return self._itemFactories

    def getItemFactories(self, return_disabled=False):
        """
        returns the list of item factories entry points currently in use

        :param return_disabled: If False (default), it returns only a list of
                                the enabled factories. If True, it returns a
                                tuple containing two lists: the enabled and
                                the available but disabled factories.
        """
        enabled = self._itemFactories
        if return_disabled:
            all_ = selectEntryPoints('taurus.form.item_factories')
            return enabled, [f for f in all_ if f not in enabled]
        else:
            return enabled

    def _customWidgetFactory(self, model):
        """
        Taurus Value Factory to provide backwards-compatibility for code
        relaying on deprecated customWidGetMap API

        :param model: taurus model object

        :return: custom TaurusValue class
        """
        try:
            key = model.getDeviceProxy().info().dev_class
            name, args, kwargs = self._customWidgetMap[key]
            pkgname, klassname = name.rsplit('.', 1)
            pkg = __import__(pkgname, fromlist=[klassname])
            klass = getattr(pkg, klassname)
        except Exception:
            return None
        return klass(*args, **kwargs)

    def _splitModel(self, modelNames):
        '''convert str to list if needed (commas and whitespace are considered as separators)'''
        if isinstance(modelNames, binary_type):
            modelNames = modelNames.decode()
        if isinstance(modelNames, string_types):
            modelNames = str(modelNames).replace(',', ' ')
            modelNames = modelNames.split()
        return modelNames

    @deprecation_decorator(alt="item factories", rel="4.6.5")
    def setCustomWidgetMap(self, cwmap):
        '''Sets a map map for custom widgets.

        :param cwmap: (dict<str,tuple>) a dictionary whose keys are device
                      type strings (i.e. see :class:`PyTango.DeviceInfo`) and
                      whose values are tuples of classname,args,kwargs
        '''
        # TODO: tango-centric
        self._customWidgetMap = cwmap

    @deprecation_decorator(alt="item factories", rel="4.6.5")
    def getCustomWidgetMap(self):
        '''Returns the map used to create custom widgets.

        :return: (dict<str,tuple>) a dictionary whose keys are device
                 type strings (i.e. see :class:`PyTango.DeviceInfo`) and whose
                 values are tuples of classname,args,kwargs
        '''
        # TODO: tango-centric
        return self._customWidgetMap

    @Qt.pyqtSlot('QString', name='modelChanged')
    def parentModelChanged(self, parentmodel_name):
        self.info("Parent model changed to '%s'" % parentmodel_name)
        parentmodel_name = str(parentmodel_name)
        if self.getUseParentModel():
            # reset the model of childs
            for obj, model in zip(self.getItems(), self.getModel()):
                obj.setModel('%s/%s' % (parentmodel_name, str(model)))
        else:
            self.debug(
                "received event from parent although not using parent model")

    def chooseModels(self):
        '''launches a model chooser dialog to modify the contents of the form'''
        if self.__modelChooserDlg is None:
            self.__modelChooserDlg = Qt.QDialog(self)
            self.__modelChooserDlg.setWindowTitle(
                "%s - Model Chooser" % str(self.windowTitle()))
            self.__modelChooserDlg.modelChooser = TaurusModelChooser()
            layout = Qt.QVBoxLayout()
            layout.addWidget(self.__modelChooserDlg.modelChooser)
            self.__modelChooserDlg.setLayout(layout)
            self.__modelChooserDlg.modelChooser.updateModels.connect(self.setModel)

        models_and_labels = []
        indexdict = {}
        for m in self.getModel():
            key = _normalize_model_name_case(m)
            indexdict[key] = indexdict.get(key, -1) + 1
            item = self.getItemByModel(m, indexdict[key])
            if item is None:
                label = None
            else:
                try:
                    # this assumes that the readwidget is a QLabel subclass (or
                    # something that has text())
                    label = str(item.labelWidget().text())
                except:
                    label = None
            models_and_labels.append((m, label))

        self.__modelChooserDlg.modelChooser.setListedModels(models_and_labels)
        self.__modelChooserDlg.show()
        self.__modelChooserDlg.raise_()

    def chooseAttrs(self):
        self.info(
            'TaurusForm.chooseAttrs() ahs been deprecated. Use TaurusForm.chooseModels() instead')
        self.chooseModels()

    def sizeHint(self):
        return Qt.QWidget.sizeHint(self)

    def _connectButtons(self):
        self.buttonBox.applyClicked.connect(self.apply)
        self.buttonBox.resetClicked.connect(self.reset)

    def getModel(self):
        return self._model

    @Qt.pyqtSlot('QStringList')
    def addModels(self, modelNames):
        '''Adds models to the existing ones:

        :param modelNames:  (sequence<str>) the names of the models to be added

        .. seealso:: :meth:`removeModels`
        '''
        modelNames = self._splitModel(modelNames)
        self.setModel(self.getModel() + modelNames)

    @Qt.pyqtSlot('QStringList')
    def removeModels(self, modelNames):
        '''Removes models from those already in the form.

        :param modelNames:  (sequence<str>) the names of the models to be removed

        .. seealso:: :meth:`addModels`
        '''
        modelNames = self._splitModel(modelNames)
        currentModels = self.getModel()
        for name in modelNames:
            try:
                currentModels.remove(name)
            except:
                self.warning("'%s' not in model list" % name)
        self.setModel(currentModels)

    def setModelCheck(self, model, check=True):
        if model is None:
            model = []
        model = [str(m or '') for m in self._splitModel(model)]
        self.destroyChildren()
        self._model = model
        self.fillWithChildren()
        # update the modelchooser list
        if self.__modelChooserDlg is not None:
            self.__modelChooserDlg.modelChooser.setListedModels(self._model)

    def resetModel(self):
        self.destroyChildren()
        self._model = []

    @deprecation_decorator(alt="item factories", rel="4.6.5")
    def getFormWidget(self, model=None):
        '''Returns a tuple that can be used for creating a widget for a given model.

        :param model: (str) a taurus model name for which the new item of the
                      form will be created

        :return: (tuple<type,list,dict>) a tuple containing a class, a list of
                 args and a dict of keyword args. The args and the keyword args
                 can be passed to the class constructor
        '''
        if model is None:
            return self._defaultFormWidget, (), {}
        # If a model is given, check if is in the custom widget map
        try:
            # if it is not an attribute, it will get an exception here
            obj = taurus.Attribute(model)
            return self._defaultFormWidget, (), {}
        except:
            try:
                obj = taurus.Device(model)
            except Exception as e:
                self.warning(
                    'Cannot handle model "%s". Using default widget.', model)
                self.debug('Model error: %s', e)
                return self._defaultFormWidget, (), {}
            try:
                key = obj.getDeviceProxy().info().dev_class  # TODO: Tango-centric
            except:
                return self._defaultFormWidget, (), {}
            #value = self._formWidgetsMap.get(key, self._defaultFormWidget)
            cwmap = self.getCustomWidgetMap()
            value = cwmap.get(key, self._defaultFormWidget)
            if isinstance(value, type):  # for backwards compatibility
                if issubclass(value, self._defaultFormWidget):
                    return value, (), {}
                else:
                    return self._defaultFormWidget, (), {'customWidgetMap': {key: value}}
            # we expect a tuple of str,list,dict -->
            # (classname_including_full_module, args, kwargs)
            name, args, kwargs = value
            pkgname, klassname = name.rsplit('.', 1)
            try:
                pkg = __import__(pkgname, fromlist=[klassname])
                klass = getattr(pkg, klassname)
            except:
                self.warning(
                    'Cannot import "%s". Using default widget for "%s".' % (name, model))
                return self._defaultFormWidget, (), {}
            if not issubclass(klass, self._defaultFormWidget):
                cwmap = kwargs.get('customWidgetMap', {})
                cwmap.update({key: klass})
                kwargs['customWidgetMap'] = cwmap
                klass = self._defaultFormWidget
            return klass, args, kwargs

    @deprecation_decorator(alt="item factories", rel="4.6.5")
    def setFormWidget(self, formWidget):
        if formWidget is None:
            from taurus.qt.qtgui.panel import TaurusValue
            formWidget = TaurusValue
        self._defaultFormWidget = formWidget

    @deprecation_decorator(alt="item factories", rel="4.6.5")
    def resetFormWidget(self):
        self.setFormWidget(self, None)

    def isWithButtons(self):
        return self._withButtons

    def setWithButtons(self, trueFalse):
        self._withButtons = trueFalse
        self.buttonBox.setVisible(self._withButtons)
        self.showButtonsAction.setChecked(self._withButtons)

    def resetWithButtons(self):
        self.setWithButtons(True)

    def onSetFormatter(self):
        """Reimplemented from TaurusBaseWidget"""
        # Form delegates se to the taurusvalues
        format = TaurusWidget.onSetFormatter(self)
        if format is not None:
            for item in self.getItems():
                rw = item.readWidget(followCompact=True)
                if hasattr(rw, 'setFormat'):
                    rw.setFormat(format)
        return format

    def setFormat(self, format):
        """
        Reimplemented to call setFormat on the taurusvalues
        """
        TaurusWidget.setFormat(self, format)
        for item in self.getItems():
            if hasattr(item, 'setFormat'):
                item.setFormat(format)

    def setCompact(self, compact):
        self._compact = compact
        for item in self.getItems():
            item.setCompact(compact)
        self.compactModeAction.setChecked(compact)

    def isCompact(self):
        return self._compact

    def resetCompact(self):
        from taurus import tauruscustomsettings
        self.setCompact(getattr(tauruscustomsettings, 'T_FORM_COMPACT', {}))

    def dropEvent(self, event):
        '''reimplemented to support dropping of modelnames in forms'''
        mtype = self.handleMimeData(event.mimeData(), self.addModels)
        if mtype is None:
            self.info('Invalid model in dropped data')
        else:
            event.acceptProposedAction()

    def setModifiableByUser(self, modifiable):
        '''
        sets whether the user can change the contents of the form
        (e.g., via Modify Contents in the context menu)
        Reimplemented from :meth:`TaurusWidget.setModifiableByUser`

        :param modifiable: (bool)

        .. seealso:: :meth:`TaurusWidget.setModifiableByUser`
        '''
        TaurusWidget.setModifiableByUser(self, modifiable)
        self.chooseModelsAction.setEnabled(modifiable)
        self.showButtonsAction.setEnabled(modifiable)
        self.changeLabelsAction.setEnabled(modifiable)
        self.compactModeAction.setEnabled(modifiable)
        for item in self.getItems():
            try:
                item.setModifiableByUser(modifiable)
            except:
                pass

    def setRegExp(self, regExp):
        pass

    def getRegExp(self):
        pass

    def destroyChildren(self):
        for child in self._children:
            self.unregisterConfigurableItem(child)
            # child.destroy()
            child.setModel(None)
            child.deleteLater()
        self._children = []

    def fillWithChildren(self):
        frame = TaurusWidget()
        frame.setLayout(Qt.QGridLayout())
        frame.layout().addItem(Qt.QSpacerItem(
            0, 0, Qt.QSizePolicy.Minimum, Qt.QSizePolicy.MinimumExpanding))

        parent_name = None
        if self.getUseParentModel():
            parent_model = self.getParentModelObj()
            if parent_model:
                parent_name = parent_model.getFullName()

        for i, model in enumerate(self.getModel()):
            if not model:
                continue
            try:
                model_obj = taurus.Object(model)
            except:
                self.warning('problem adding item "%s"', model)
                self.traceback(level=taurus.Debug)
                continue
            if parent_name:
                # @todo: Change this (it assumes tango model naming!)
                model = "%s/%s" % (parent_name, model)

            widget = None
            # check if some item factory handles this model
            for ep in self._itemFactories:
                try:
                    # load the plugin
                    f = ep.load()
                except:
                    warning('cannot load item factory "%s"', ep.name)
                    continue
                try:
                    widget = f(model_obj)
                except Exception as e:
                    warning('factory "%s" raised "%r" for "%s". '
                            + 'Tip: consider disabling it', ep.name, e, model)
                if widget is not None:
                    self.debug(
                        "widget for '%s' provided by '%s'",
                        model,
                        ep.name
                    )
                    break

            # backwards-compat with deprecated custom widget map API
            if widget is None and self._customWidgetMap:
                widget = self._customWidgetFactory(model_obj)

            # no factory handles the model and no custom widgets. Use default
            if widget is None:
                widget = self._defaultFormWidget()

            # @todo UGLY... See if this can be done in other ways... (this causes trouble with widget that need more vertical space , like PoolMotorTV)
            widget.setMinimumHeight(20)

            try:
                widget.setCompact(self.isCompact())
                widget.setModel(model)
                widget.setParent(frame)
            except:
                # raise
                self.warning('problem adding item "%s"', model)
                self.traceback(level=taurus.Debug)
            try:
                widget.setModifiableByUser(self.isModifiableByUser())
            except:
                pass
            try:
                widget.setFormat(self.getFormat())
            except Exception:
                self.debug('Cannot set format %s to child %s',
                           self.getFormat(), model)
            widget.setObjectName("__item%i" % i)
            self.registerConfigDelegate(widget)
            self._children.append(widget)

        frame.layout().addItem(Qt.QSpacerItem(
            0, 0, Qt.QSizePolicy.Minimum, Qt.QSizePolicy.MinimumExpanding))
        self.scrollArea.setWidget(frame)
#        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setMinimumWidth(frame.layout().sizeHint().width() + 20)

    def getItemByModel(self, model, index=0):
        '''returns the child item with given model. If there is more than one item
        with the same model, the index parameter can be used to distinguish among them
        Please note that his index is only relative to same-model items!'''
        for child in self._children:
            if (_normalize_model_name_case(child.getModel()) ==
                    _normalize_model_name_case(model)):
                if index <= 0:
                    return child
                else:
                    index -= 1

    def getItemByIndex(self, index):
        '''returns the child item with at the given index position. '''
        return self.getItems()[index]

    def getItems(self):
        '''returns a list of the objects that have been created as childs of the form'''
        return self._children

#    def _manageButtonBox(self):
#        if self.isWithButtons():
#            self.buttonBox.setVisible(True)
#        else:
#            self.buttonBox.setVisible(False)

    def onChangeLabelsAction(self):
        '''changes the labelConfig of all its items'''
        keys = ['{attr.label}', '{attr.name}', '{attr.fullname}', '{dev.name}',
                '{dev.fullname}']

        msg = 'Choose the label format. \n' + \
              'You may use Python format() syntax. The TaurusDevice object\n' + \
              'can be referenced as "dev" and the TaurusAttribute object\n' + \
              'as "attr"'
        labelConfig, ok = Qt.QInputDialog.getItem(self, 'Change Label', msg,
                                                  keys, 0, True)
        if ok:
            for item in self.getItems():
                item.labelConfig = (str(labelConfig))

    @Qt.pyqtSlot()
    def apply(self):
        self.safeApplyOperations()

    @Qt.pyqtSlot()
    def reset(self):
        self.resetPendingOperations()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.panel'
        ret['container'] = False
        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QStringList", getModel,
                            TaurusWidget.setModel,
                            resetModel)

    useParentModel = Qt.pyqtProperty("bool",
                                     TaurusWidget.getUseParentModel,
                                     TaurusWidget.setUseParentModel,
                                     TaurusWidget.resetUseParentModel)

    withButtons = Qt.pyqtProperty("bool", isWithButtons,
                                  setWithButtons,
                                  resetWithButtons)

    modifiableByUser = Qt.pyqtProperty("bool", TaurusWidget.isModifiableByUser,
                                       setModifiableByUser,
                                       TaurusWidget.resetModifiableByUser)

    compact = Qt.pyqtProperty("bool", isCompact, setCompact, resetCompact)


class TaurusCommandsForm(TaurusWidget):
    '''A form that shows commands available for a Device Server'''

    # TODO: tango-centric
    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)

        self.setLayout(Qt.QVBoxLayout())

        self._splitter = Qt.QSplitter()
        self._splitter.setOrientation(Qt.Qt.Vertical)
        self.layout().addWidget(self._splitter)

        self._frame = TaurusWidget(self)
        self._frame.setLayout(Qt.QGridLayout())

        self._scrollArea = TaurusScrollArea(self)
        self._scrollArea.setWidget(self._frame)
        self._scrollArea.setWidgetResizable(True)
        self._splitter.addWidget(self._scrollArea)

        self._outputTE = Qt.QTextEdit()
        self._outputTE.setReadOnly(True)
        self._splitter.addWidget(self._outputTE)
        #self._splitter.moveSplitter(self._splitter.getRange(0)[-1], 0)

        self._cmdWidgets = []
        self._paramWidgets = []
        self._viewFilters = []
        self._defaultParameters = []
        self._sortKey = lambda x: x.cmd_name

        self._operatorViewFilter = lambda x: x.disp_level == DisplayLevel.OPERATOR

        # self.setLayout(Qt.QGridLayout())
        self.modelChanged.connect(self._updateCommandWidgets)

    def createConfig(self, allowUnpickable=False):
        '''
        extending  :meth:`TaurusBaseWidget.createConfig`
        :param alllowUnpickable:  (bool)

        :return: (dict<str,object>) configurations (which can be loaded with :meth:`applyConfig`).

        .. seealso: :meth:`TaurusBaseWidget.createConfig`, :meth:`applyConfig`
        '''
        # get the basic config
        configdict = TaurusWidget.createConfig(
            self, allowUnpickable=allowUnpickable)
        # store the splitter config
        configdict['splitter/state'] = self._splitter.saveState().data()
        return configdict

    def applyConfig(self, configdict, **kwargs):
        """extending :meth:`TaurusBaseWidget.applyConfig` to restore the splitter config

        :param configdict: (dict)

        .. seealso:: :meth:`TaurusBaseWidget.applyConfig`, :meth:`createConfig`
        """
        # first do the basic stuff...
        TaurusWidget.applyConfig(self, configdict, **kwargs)
        # restore the splitter config
        self._splitter.restoreState(
            Qt.QByteArray(configdict['splitter/state']))

    def getModelClass(self):
        '''see :meth:`TaurusBaseComponent.getModelClass`'''
        return taurus.core.taurusdevice.TaurusDevice

    def _updateCommandWidgets(self, *args):
        '''
        Inserts command buttons and parameter widgets in the layout, according to
        the commands from the model
        '''

        dev = self.getModelObj()
        if dev is None:
            self._clearFrame()
            return

        try:
            commands = sorted(dev.command_list_query(), key=self._sortKey)
        except Exception as e:
            self.warning('Problem querying commands from %s. Reason: %s',
                         dev, e)
            self._clearFrame()
            return

        for f in self.getViewFilters():
            commands = list(filter(f, commands))

        self._clearFrame()

        layout = self._frame.layout()

        model = self.getFullModelName()

        for row, c in enumerate(commands):
            self.debug('Adding button for command %s' % c.cmd_name)
            button = TaurusCommandButton(command=c.cmd_name, text=c.cmd_name)
            layout.addWidget(button, row, 0)
            button.setModel(model)
            self._cmdWidgets.append(button)
            button.commandExecuted.connect(self._onCommandExecuted)
            
            import taurus.core.tango.util.tango_taurus as tango_taurus
            in_type = tango_taurus.FROM_TANGO_TO_TAURUS_TYPE[c.in_type]

            if in_type is not None:
                self.debug('Adding arguments for command %s' % c.cmd_name)
                pwidget = ParameterCB()
                if c.cmd_name.lower() in self._defaultParameters:
                    for par in self._defaultParameters.get(c.cmd_name.lower(), []):
                        pwidget.addItem(par)
                    #pwidget.setEditable( (self._defaultParameters[c.cmd_name.lower()] or [''])[0] == '' )
                    if (self._defaultParameters[c.cmd_name.lower()] or [''])[0] == '':
                        pwidget.setEditable(True)
                    else:
                        pwidget.setEditable(False)
                        button.setParameters(self._defaultParameters[
                                             c.cmd_name.lower()][0])
                pwidget.editTextChanged.connect(button.setParameters)
                pwidget.currentIndexChanged['QString'].connect(button.setParameters)
                pwidget.activated.connect(button.setFocus)
                button.commandExecuted.connect(pwidget.rememberCurrentText)
                layout.addWidget(pwidget, row, 1)
                self._paramWidgets.append(pwidget)

    def _clearFrame(self):
        '''destroys all widgets in the scroll area frame'''
        self._frame = TaurusWidget(self)
        self._frame.setLayout(Qt.QGridLayout())
        # hack because useParentModel does not work!
        self._frame.setModel(self.getModelName())
        self._scrollArea.setWidget(self._frame)
        self._cmdWidgets = []
        self._paramWidgets = []

    def _onCommandExecuted(self, result):
        '''Slot called when the command is executed, to manage the output

        :param result: return value from the command. The type depends on the command
        '''
        timestamp = datetime.now()
        cmdbutton = self.sender()
        output = '<i>%s</i><br>' % timestamp.strftime('%Y-%m-%d %H:%M:%S') +\
                 '<b>Command:</b> %s<br>' % cmdbutton.getCommand() +\
                 '<b>Pars:</b> %s<br>' % repr(cmdbutton.getParameters()) +\
                 '<b>Return Value:</b><br>%s' % str(result)
        self._outputTE.append(output)
        # ugly workaround for bug in html rendering in Qt
        separator = '<table width=\"100%\"><tr><td><hr /></td></tr></table>'
        # see http://lists.trolltech.com/qt-interest/2008-04/thread00224-0.html
        self._outputTE.append(separator)
        # self._outputTE.append('%s'%timestamp)
        #self._outputTE.append('<b>Command:</b> "%s"'%cmdbutton.getCommand())
        #self._outputTE.append('<b>Pars:</b> %s'%repr(cmdbutton.getParameters()))
        #self._outputTE.append('<b>Return Value:</b><br>%s<hr>'%str(result))

    def setSortKey(self, sortkey):
        '''sets the method used to sort the commands (cmd_name by default)

        :param sortkey: (callable) a function that takes a :class:`CommandInfo`
                        as argument and returns a key to use for sorting purposes
                        (e.g. the default sortKey is ``lambda x:x.cmd_name``)
        '''
        self._sortKey = sortkey
        self._updateCommandWidgets()

    def setDefaultParameters(self, params):
        '''sets the values that will appear by default in the parameters combo box,
        the command combo box for the command will be editable only if the first parameter is an empty string

        :param params: (dict<str,list>) { 'cmd_name': ['parameters string 1', 'parameters string 2' ] }

        '''
        self._defaultParameters = dict((k.lower(), v)
                                       for k, v in params.items())
        self._updateCommandWidgets()

    def setViewFilters(self, filterlist):
        '''sets the filters to be applied when displaying the commands

        :param filterlist: (sequence<callable>) a sequence of command filters.
                           All filters will be applied to each command to decide
                           whether to display it or not.
                           for a command to be plotted, the following condition
                           must be true for all filters:
                           bool(filter(command))==True
        '''
        self._viewFilters = filterlist
        self._updateCommandWidgets()

    def getViewFilters(self):
        '''returns the filters used in deciding which commands are displayed

        :return: (sequence<callable>) a sequence of filters
        '''
        return self._viewFilters

    @Qt.pyqtSlot(bool, name="setCommand")
    def setExpertView(self, expert):
        '''sets the expert view mode

        :param expert: (bool) If expert is True, commands won't be filtered. If
                       it is False, commands with  display level Expert won't be shown
        '''
        currentfilters = self.getViewFilters()
        if expert:
            if self._operatorViewFilter in currentfilters:
                currentfilters.remove(self._operatorViewFilter)
        else:
            if self._operatorViewFilter not in currentfilters:
                currentfilters.insert(0, self._operatorViewFilter)
        self.setViewFilters(currentfilters)
        self._expertView = expert

    def getSplitter(self):
        '''returns the splitter that separates the command buttons area from
        the output area

        :return: (QSplitter)'''
        return self._splitter

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.panel'
        ret['container'] = False
        return ret


class TaurusAttrForm(TaurusWidget):
    '''A form that displays the attributes of a Device Server'''

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)

        self._viewFilters = []
        self._operatorViewFilter = lambda x: x.disp_level == DisplayLevel.OPERATOR

        self.setLayout(Qt.QVBoxLayout())

        self._form = TaurusForm(parent=self)
        self.layout().addWidget(self._form)
        self.registerConfigDelegate(self._form)

        self.modelChanged.connect(self._updateAttrWidgets)

        self._sortKey = lambda x: x.name

    def setSortKey(self, sortkey):
        '''sets the key function used to sort the attributes in the form

        :param sortkey: (callable) a function that takes a :class:`AttributeInfo`
                        as argument and returns a key to use for sorting purposes
                        (e.g. the default sortKey is ``lambda x:x.name``)

        '''
        self._sortKey = sortkey
        self._updateAttrWidgets()

    def getModelClass(self):
        '''see :meth:`TaurusBaseComponent.getModelClass`'''
        return taurus.core.taurusdevice.TaurusDevice

    def _updateAttrWidgets(self):
        '''Populates the form with an item for each of the attributes shown
        '''
        try:
            dev = self.getModelObj()
            attrlist = sorted(dev.attribute_list_query(), key=self._sortKey)
            for f in self.getViewFilters():
                attrlist = list(filter(f, attrlist))
            attrnames = []
            devname = self.getModelName()
            for a in attrlist:
                # ugly hack . But setUseParentModel does not work well
                attrnames.append("%s/%s" % (devname, a.name))
            self.debug('Filling with attribute list: %s'
                       % ("; ".join(attrnames)))
            self._form.setModel(attrnames)
        except:
            self.debug('Cannot connect to device')
            self._form.setModel([])

    def setViewFilters(self, filterlist):
        '''sets the filters to be applied when displaying the attributes

        :param filterlist: (sequence<callable>) a sequence of attr filters. All
                           filters will be applied to each attribute name to
                           decide whether to display it or not. for an attribute
                           to be plotted, the following condition must be true
                           for all filters: ``bool(filter(attrname))==True``
        '''
        self._viewFilters = filterlist
        self._updateAttrWidgets()

    def getViewFilters(self):
        '''returns the filters used in deciding which attributes are displayed

        :return: (sequence<callable>) a sequence of filters
        '''
        return self._viewFilters

    @Qt.pyqtSlot(bool)
    def setExpertView(self, expert):
        '''sets the expert view mode

        :param expert: (bool) If expert is True, attributes won't be filtered. If
                       it is False, attributes with display level Expert won't
                       be shown
        '''
        currentfilters = self.getViewFilters()
        if expert:
            if self._operatorViewFilter in currentfilters:
                currentfilters.remove(self._operatorViewFilter)
        else:
            if self._operatorViewFilter not in currentfilters:
                currentfilters.insert(0, self._operatorViewFilter)
        self.setViewFilters(currentfilters)
        self._expertView = expert

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.panel'
        ret['container'] = False
        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", TaurusWidget.getModel,
                            TaurusWidget.setModel,
                            TaurusWidget.resetModel)

    useParentModel = Qt.pyqtProperty("bool",
                                     TaurusWidget.getUseParentModel,
                                     TaurusWidget.setUseParentModel,
                                     TaurusWidget.resetUseParentModel)


def test1():
    '''tests taurusForm'''
    import sys
    if len(sys.argv) > 1:
        models = sys.argv[1:]
    else:
        models = None
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv, cmd_line_parser=None)
    if models is None:
        models = ['sys/tg_test/1/state',
                  'sys/tg_test/1/float_scalar',
                  'sys/tg_test/1/boolean_image',
                  'sys/tg_test/1/float_spectrum',
                  'sys/tg_test/1/status']
    dialog = TaurusForm()
    dialog.setModel(models)
    dialog.setModifiableByUser(True)
    for i, tv in enumerate(dialog.getItems()):
        tv.setDangerMessage("Booooo scaring %d!!!" % i)
    dialog.show()

    dialog2 = TaurusForm()
    dialog2.show()
    dialog2.setModifiableByUser(True)
    sys.exit(app.exec_())


def test2():
    '''tests taurusAttrForm'''
    import sys
    if len(sys.argv) > 1:
        model = sys.argv[1]
    else:
        model = None

    if model is None:
        model = 'bl97/pc/dummy-01'
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv, cmd_line_parser=None)
    dialog = TaurusAttrForm()
    dialog.setModel(model)
    dialog.show()
    sys.exit(app.exec_())


def test3():
    '''tests taurusCommandsForm'''
    import sys
    if len(sys.argv) > 1:
        model = sys.argv[1]
    else:
        model = None

    if model is None:
        model = 'bl97/pc/dummy-01'
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv, cmd_line_parser=None)
    dialog = TaurusCommandsForm()
    dialog.setModel(model)
    dialog.show()
#    dialog.getSplitter().setSizes([10,220])

    sys.exit(app.exec_())


@click.command('form')
@taurus.cli.common.window_name("TaurusForm")
@taurus.cli.common.config_file
@click.option(
    '--inc-fact',
    multiple=True,
    metavar="INC",
    default=getattr(_ts, 'T_FORM_ITEM_FACTORIES', {}).get('include', ('.*',)),
    show_default=True,
    type=click.STRING,
    help=("Enable item factories matching INC pattern"
          + " (can be passed multiple times)")
)
@click.option(
    '--exc-fact',
    multiple=True,
    metavar="EXC",
    default=getattr(_ts, 'T_FORM_ITEM_FACTORIES', {}).get('exclude', ()),
    show_default=True,
    type=click.STRING,
    help=("Disable item factories matching EXC pattern"
          + " (can be passed multiple times)")
)
@click.option(
    '--ls-fact',
    is_flag=True,
    help="List the available item factories"
)
@taurus.cli.common.models
def form_cmd(window_name, config_file, inc_fact, exc_fact, ls_fact, models):
    """Shows a Taurus form populated with the given model names"""
    from taurus.qt.qtgui.application import TaurusApplication
    import sys
    app = TaurusApplication(cmd_line_parser=None)
    dialog = TaurusForm()

    dialog.setItemFactories(include=inc_fact, exclude=exc_fact)

    if ls_fact:
        inc, exc = dialog.getItemFactories(return_disabled=True)
        msg = "\nItem Factories in {}:\n".format(window_name)
        msg += "\n".join(["  [*] " + e.name for e in inc] +
                         ["  [ ] " + e.name for e in exc])
        msg += "\nPatterns used for item factory selection:\n"
        msg += "  INC: {}\n".format(inc_fact)
        msg += "  EXC: {}\n".format(exc_fact)
        print(msg)
        click.get_current_context().exit(0)

    dialog.setModifiableByUser(True)
    dialog.setModelInConfig(True)

    dialog.setWindowTitle(window_name)

    # Make sure the window size and position are restored
    dialog.registerConfigProperty(dialog.saveGeometry, dialog.restoreGeometry,
                                  'MainWindowGeometry')

    quitApplicationAction = Qt.QAction(
        Qt.QIcon.fromTheme("process-stop"), 'Close Form', dialog)
    quitApplicationAction.triggered.connect(dialog.close)

    saveConfigAction = Qt.QAction("Save current settings...", dialog)
    saveConfigAction.setShortcut(Qt.QKeySequence.Save)
    saveConfigAction.triggered.connect(
        partial(dialog.saveConfigFile, ofile=None))
    loadConfigAction = Qt.QAction("&Retrieve saved settings...", dialog)
    loadConfigAction.setShortcut(Qt.QKeySequence.Open)
    loadConfigAction.triggered.connect(
        partial(dialog.loadConfigFile, ifile=None))

    dialog.addActions(
        (saveConfigAction, loadConfigAction, quitApplicationAction))

    # backwards-compat: in case T_FORM_CUSTOM_WIDGET_MAP was manually edited
    from taurus import tauruscustomsettings
    cwmap = getattr(tauruscustomsettings, 'T_FORM_CUSTOM_WIDGET_MAP', {})
    if cwmap:
        dialog.setCustomWidgetMap(cwmap)

    # set a model list from the command line or launch the chooser
    if config_file is not None:
        dialog.loadConfigFile(config_file)
    elif len(models) > 0:
        dialog.setModel(models)
    else:
        dialog.chooseModels()

    dialog.show()
    sys.exit(app.exec_())


def main():
    # test1()
    # test2()
    # test3()
    # test4()
    form_cmd()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

"""
This package provides TaurusGui, a generic framework for creating GUIs without
actual coding (just configuration files).

See the examples provided in the conf subdirectory directory as well as the
documentation of the :class:`TaurusGui` class.


The "new GUI wizard" and XML configuration files
------------------------------------------------

Note that the configuration files can either be written by hand or by
launching the "new GUI" wizard with `taurus newgui`, which will create
a new directory containing configuration, resource and launcher files.

The new GUI wizard stores all the options in xml format in a file called
`config.xml` and creates a simple `config.py` file containing the following
line::

    XML_CONFIG = 'config.xml'

This line indicates that `config.xml` should also be used as a source of
configuration options (in case of conflict, the options set in `config.py`
prevail).

"""
from __future__ import absolute_import

from .paneldescriptionwizard import *
from .taurusgui import *
from .appsettingswizard import *
try:
    # this import is left here for bck-compat, but will be removed
    # TODO: remove this
    from sardana.taurus.qt.qtgui.macrolistener import (MacroBroker,
                                                       DynamicPlotManager)
except ImportError:
    pass


__docformat__ = 'restructuredtext'



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Release data for the taurus project. It contains the following members:

    - version : (str) version string
    - description : (str) brief description
    - long_description : (str) a long description
    - license : (str) license
    - authors : (dict<str, tuple<str,str>>) the list of authors
    - url : (str) the project url
    - download_url : (str) the project download url
    - platforms : list<str> list of supported platforms
    - keywords : list<str> list of keywords
"""

# -*- coding: utf-8 -*-
__docformat__ = "restructuredtext"


# Name of the package for release purposes.  This is the name which labels
# the tarballs and RPMs made by distutils, so it's best to lowercase it.
name = 'taurus'

# we use semantic versioning (http://semver.org/) and we update it using the
# bumpversion script (https://github.com/peritus/bumpversion)
version = '4.7.1'

# generate version_info and revision (**deprecated** since version 4.0.2-dev).
if '-' in version:
    _v, _rel = version.split('-')
else:
    _v, _rel = version, ''
_v = tuple([int(n) for n in _v.split('.')])
version_info = _v + (_rel, 0)   # deprecated, do not use
revision = str(version_info[4])  # deprecated, do not use


description = "A framework for scientific/industrial CLIs and GUIs"

long_description = """Taurus is a python framework for control and data
acquisition CLIs and GUIs in scientific/industrial environments.
It supports multiple control systems or data sources: Tango, EPICS,...
New control system libraries can be integrated through plugins."""

license = 'LGPL'

authors = {'Tiago_et_al': ('Tiago Coutinho et al.', ''),
           'Community': ('Taurus Community',
                         'tauruslib-devel@lists.sourceforge.net'),
           }


url = 'http://www.taurus-scada.org'

download_url = 'http://pypi.python.org/packages/source/t/taurus'

platforms = ['Linux', 'Windows']

keywords = ['CLI', 'GUI', 'PyTango', 'Tango', 'Shell', 'Epics']
